/*
 * Copyright (c) 2005,2006 Ryo Ichinose
 *
 * This file is part of Space Maze.
 * 
 * Space Maze is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; version 2 of the License.
 * 
 * Space Maze is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with Space Maze; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

/*!
 * @file
 * 
 * Game manager recognizes the current game status, and
 * dispatches process according to the current game status 
 * within updating and drawing function called within 
 * main loop.
 *
 * "$Id: game_manager.c,v 1.2 2006/07/23 15:10:50 ryo Exp $"
 */

#include "RLgt.h"
#include "spacemaze.h"
#include "state.h"


/*!
 * Initializes game manager.
 * @param app [i/o] Application context.
 */
void
game_manager_init( SpaceMazeApp* app )
{
	if ( app == NULL ) {
		return;
	}

	INIT_STARS( app );

	app->game_state = STATE_TITLE;
	app->next_state = STATE_INVALID;
	state_title_init( app );
}


typedef struct
{
	int  ( *update_func )( SpaceMazeApp* );
	void ( *draw_func )( SpaceMazeApp* );
}
StateFuncMap;


static StateFuncMap state_func_map[] =
{
	{ state_title_update,    state_title_draw    },
	{ state_select_update,   state_select_draw   },
	{ state_boss_update,     state_boss_draw     },
	{ state_playing_update,  state_playing_draw  },
	{ state_pause_update,    state_pause_draw    },
	{ state_clear_update,    state_clear_draw    },
	{ state_gameover_update, state_gameover_draw },
	{ state_ending_update,   state_ending_draw   }
};


/*!
 * Updates internal information of all modules.
 * This function will be called within main loop at each frame.
 * @param app [i/o] Application context.
 */
void
game_manager_update( SpaceMazeApp* app )
{
	if ( app == NULL ) {
		return;
	}

	app->next_state = state_func_map[ app->game_state ].update_func( app );

	if ( app->next_state != app->game_state ) {
		app->game_state = app->next_state;
		app->next_state = STATE_INVALID;
	}
}


/*!
 * Draws screen according to the current game status.
 * @param app [i/o] Application context.
 */
void
game_manager_draw( SpaceMazeApp* app )
{
	if ( app == NULL || app->next_state == STATE_INVALID ) {
		return;
	}

	state_func_map[ app->game_state ].draw_func( app );
}


/* End of game_manager.c */
