/*
 * Copyright (c) 2006 Ryo Ichinose
 *
 * This file is part of Space Maze.
 * 
 * Space Maze is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; version 2 of the License.
 * 
 * Space Maze is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with Space Maze; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

/*!
 * @file
 * @brief Handles enemy (flyer) data and draws it onto the screen.
 *
 * "$Id: flyer.c,v 1.2 2006/09/21 14:45:19 ryo Exp $"
 */

#include <math.h>
#include <stdio.h>
#include <stdlib.h>

#include "RLgt.h"
#include "RLrect.h"
#include "RLsound.h"
#include "enemy.h"
#include "entity.h"
#include "resource.h"
#include "spaceship.h"


#define FLYER_WIDTH              40
#define FLYER_HEIGHT             40
#define FLYER_TRACE_DISTANCE     300
#define FLYER_MAX_ENERGY         150
#define FLYER_STRENGTH           15
#define FLYER_TRACE_VELOCITY     1.6
#define FLYER_NORMAL_VELOCITY    1.0

#define FLYER_MODE_NORMAL        0
#define FLYER_MODE_TRACE         1
#define FLYER_MODE_SUSPEND       2

#define FLYER_FRAME_COUNT_EFFECT_SUSPEND 300

#define FLYER_KEY_LOCKED               0
#define FLYER_KEY_RELEASED             1
#define FLYER_KEY_RELEASED_AND_NOTICED 2


static void flyer_trace( Entity* flyer, Entity* spaceship );
static void flyer_draw_body( Entity* flyer );
static void flyer_draw_circle( Entity* flyer, int screen_x, int screen_y );


/*!
 * Initializes enemy (flyer) specific data.
 * No null check is executed because this is a back-end of enemy_init().
 */
void
flyer_init( Entity* flyer )
{
	flyer->width             = FLYER_WIDTH;
	flyer->height            = FLYER_HEIGHT;
	flyer->energy            = FLYER_MAX_ENERGY;
	flyer->strength          = FLYER_STRENGTH;
	flyer->vmax              = ( FLYER_NORMAL_VELOCITY / entity_get_fps_ratio() );
	flyer->vx                = 0;
	flyer->vy                = 0;
	flyer->explode_func      = flyer_explode;
	flyer->key_released_func = flyer_key_released;
	flyer->check_func        = flyer_check;
	flyer->update_func       = flyer_update;
	flyer->draw_func         = flyer_draw;
}


/*!
 * Suspends flyer activity.
 * No null check is executed because this is a back-end of entity_explode().
 */
void
flyer_explode( Entity* flyer )
{
	if ( flyer->mode == FLYER_MODE_SUSPEND ) {
		return;
	}

	flyer->mode = FLYER_MODE_SUSPEND;
	flyer->effect_animation_frame_count = 
		( FLYER_FRAME_COUNT_EFFECT_SUSPEND * entity_get_fps_ratio() );

	if ( flyer->key_released == FLYER_KEY_LOCKED ) {
		rlib_sound_play( flyer->sound, flyer->wave_index[ INDEX_WAVE_SUSPEND ], 0 );
		flyer->key_released = FLYER_KEY_RELEASED;
	}
}


/*!
 * Checks if the key is released or not.
 * No null check is executed because this is a back-end of entity_key_released.
 */
int
flyer_key_released( Entity* flyer )
{
	if ( flyer->key_released == FLYER_KEY_RELEASED ) {
		flyer->key_released = FLYER_KEY_RELEASED_AND_NOTICED;
		return 1;
	}

	return 0;
}


/*!
 * Computes new position just for collision check.
 * Note that this function does not compute real position.
 * After this, execute entity_update() to compute real position.
 * This function returns 0 when no collision check is needed 
 * because enemy is not alive ...
 *
 * No null check is executed because this is a back-end of entity_check().
 */
int
flyer_check( Entity* flyer, RLrect* rect, double* vx, double* vy )
{
	if ( ! entity_is_alive( flyer ) ) {
		return 0;
	}

	rect->width  = ( int )( flyer->width  * 0.9 );
	rect->height = ( int )( flyer->height * 0.9 );

	if ( flyer->mode == FLYER_MODE_SUSPEND ) {
		rect->lefttop_x = flyer->px - ( flyer->width  * 0.45 );
		rect->lefttop_y = flyer->py - ( flyer->height * 0.45 );
		(*vx) = (*vy) = 0;
	}
	else {
		rect->lefttop_x = flyer->px + flyer->vx - ( flyer->width  * 0.45 );
		rect->lefttop_y = flyer->py + flyer->vy - ( flyer->height * 0.45 );
		(*vx) = flyer->vx;
		(*vy) = flyer->vy;
	}

	return 1;
}


/*!
 * Commits new position, status, etc. at the next frame.
 * No null check is executed because this is a back-end of entity_update().
 * spaceship can be NULL, which will be ignored in this function.
 */
void
flyer_update( Entity* flyer, Entity* spaceship )
{
	if ( flyer->mode != FLYER_MODE_SUSPEND ) {
		flyer->px += flyer->vx;
		flyer->py += flyer->vy;

		if ( flyer->effect_blink_frame_count > 0 ) {
			flyer->effect_blink_frame_count --;
		}
		else if ( spaceship != NULL ) {
			flyer_trace( flyer, spaceship );
		}			
	}
	else {
		flyer->effect_animation_frame_count --;

		if ( flyer->effect_animation_frame_count <= 0 ) {
			flyer->energy = FLYER_MAX_ENERGY;
			flyer->mode   = FLYER_MODE_NORMAL;
		}
	}
}


static void
flyer_trace( Entity* flyer, Entity* spaceship )
{
	double px;
	double py;
	double dx;
	double dy;
	double distance;

	entity_get_position( spaceship, &px, &py );
	distance = DISTANCE( flyer->px, flyer->py, px, py );

	if ( distance < FLYER_TRACE_DISTANCE ) {
		flyer->mode = FLYER_MODE_TRACE;
		flyer->vmax = ( FLYER_TRACE_VELOCITY / entity_get_fps_ratio() );

		dx = px - flyer->px;
		dy = py - flyer->py;
		NORMALIZE( dx, dy );
		flyer->vx = flyer->vmax * dx;
		flyer->vy = flyer->vmax * dy;
	}
	else {
		flyer->mode = FLYER_MODE_NORMAL;
		flyer->vmax = FLYER_NORMAL_VELOCITY;
		NORMALIZE( flyer->vx, flyer->vy );
		flyer->vx *= flyer->vmax;
		flyer->vy *= flyer->vmax;
	}
}


/*!
 * Draws an enemy (flyer) onto the screen.
 * No null check is executed because this is a back-end of entity_draw().
 */
void
flyer_draw( Entity* flyer )
{
	if ( entity_is_exploding( flyer ) ) {
		explosion_draw( &( flyer->explosion ) );
		return;
	}

	if ( flyer->is_alive ) {
		flyer_draw_body( flyer );
	}
}


static void
flyer_draw_body( Entity* flyer )
{
#define FLYER_BODY_POINTS_NUM 35

	const double points[ FLYER_BODY_POINTS_NUM ][2] = {
		{  10.0, -20.0 },
		{  20.0, -20.0 },
		{  20.0,  10.0 },
		{  15.0,  20.0 },
		{ -15.0,  20.0 },
		{ -20.0,  10.0 },
		{ -20.0, -20.0 },
		{ -10.0, -20.0 },
		{ -10.0, -16.6 },
		{ -15.0, -16.6 },
		{ -15.0, -13.3 },
		{ -10.0, -13.3 },
		{ -10.0, -10.0 },
		{ -15.0, -10.0 },
		{ -15.0, - 6.6 },
		{ -10.0, - 6.6 },
		{ -10.0, - 3.3 },
		{ -15.0, - 3.3 },
		{ -15.0,   0.0 },
		{ -10.0,   0.0 },
		{ - 5.0,  10.0 },
		{   5.0,  10.0 },
		{  10.0,   0.0 },
		{  15.0,   0.0 },
		{  15.0, - 3.3 },
		{  10.0, - 3.3 },
		{  10.0, - 6.6 },
		{  15.0, - 6.6 },
		{  15.0, -10.0 },
		{  10.0, -10.0 },
		{  10.0, -13.3 },
		{  15.0, -13.3 },
		{  15.0, -16.6 },
		{  10.0, -16.6 },
		{  10.0, -20.0 }
	};

	Point converted[ FLYER_BODY_POINTS_NUM ];
	int screen_x;
	int screen_y;
	int culling = 1;
	int i;

	SCREEN_POSITION( flyer, screen_x, screen_y );

	for ( i = 0; i < FLYER_BODY_POINTS_NUM; i ++ ) {
		converted[i].x = ( short )( screen_x + points[i][0] );
		converted[i].y = ( short )( screen_y + points[i][1] );

		if ( rlib_rect_within( &( flyer->screen ), converted[i].x, converted[i].y ) ) {
			culling = 0;
		}
	}

	if ( culling ) {
		return;
	}
	
	rlib_gt_set_foreground_color( flyer->gt_app, RL_GT_MAX_RGB, 0, RL_GT_MAX_RGB );
	rlib_gt_draw_lines( flyer->gt_app, converted, FLYER_BODY_POINTS_NUM );
	flyer_draw_circle( flyer, screen_x, screen_y );
}
              

static void
flyer_draw_circle( Entity* flyer, int screen_x, int screen_y )
{
#define NUM_OF_CIRCLE   3
#define DIAMETER        10

	const double lefttops[ NUM_OF_CIRCLE ][2] = {
		{ -5.0, -20.0 }, /* blue */
		{ -5.0, -10.0 }, /* yellow */
		{ -5.0,   0.0 }  /* red */
	};

	int circle_x;
	int circle_y;
	int i;

	if ( flyer->mode == FLYER_MODE_NORMAL ) {
		rlib_gt_set_foreground_color( flyer->gt_app, 0, 0, RL_GT_MAX_RGB );
	}
	else if ( flyer->mode == FLYER_MODE_TRACE ) {
		rlib_gt_set_foreground_color( flyer->gt_app, RL_GT_MAX_RGB, 0, 0 );
	}
	else if ( flyer->mode == FLYER_MODE_SUSPEND ) {
		rlib_gt_set_foreground_color( flyer->gt_app, RL_GT_MAX_RGB, RL_GT_MAX_RGB, 0 );
	}

	for ( i = 0; i < NUM_OF_CIRCLE; i ++ ) {
		circle_x = screen_x + lefttops[i][0];
		circle_y = screen_y + lefttops[i][1];
		rlib_gt_fill_arc( flyer->gt_app, circle_x, circle_y, 
				  DIAMETER, DIAMETER, 0, 360 * 64 );
	}
}


/* End of flyer.c */
