/*
 * Copyright (c) 2005,2006 Ryo Ichinose
 *
 * This file is part of Space Maze.
 * 
 * Space Maze is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; version 2 of the License.
 * 
 * Space Maze is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with Space Maze; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

/*!
 * @file
 * @brief Handles enemy (fortress) data and draws it onto the screen.
 *
 * "$Id: fortress.c,v 1.1.1.1 2006/07/22 18:07:56 ryo Exp $"
 */

#include <math.h>
#include <stdio.h>
#include <stdlib.h>

#include "RLgt.h"
#include "RLrect.h"
#include "RLsound.h"
#include "enemy.h"
#include "entity.h"
#include "explosion.h"
#include "missile.h"
#include "spaceship.h"
#include "resource.h"


#define FRAME_COUNT_TO_NEXT_LAUNCH  150

#define FORTRESS_WIDTH              300
#define FORTRESS_HEIGHT             120
#define FORTRESS_LAUNCH_DISTANCE_S  500
#define FORTRESS_LAUNCH_DISTANCE_L  1000
#define FORTRESS_MAX_ENERGY         500
#define FORTRESS_STRENGTH           20
#define FORTRESS_VELOCITY           2.0


static void fortress_draw_body( Entity* fortress );
static void fortress_launch_missile( Entity* fortress, Entity* spaceship );
static void fortress_draw_circle( Entity* fortress, int screen_x, int screen_y, int index_offset );


/*!
 * Initializes enemy (fortress) specific data.
 * No null check is executed because this is a back-end of enemy_init().
 */
void
fortress_init( Entity* fortress )
{
	double angle = RANDOM( 2.0 * PI );

	fortress->width        = FORTRESS_WIDTH;
	fortress->height       = FORTRESS_HEIGHT;
	fortress->energy       = FORTRESS_MAX_ENERGY;
	fortress->strength     = FORTRESS_STRENGTH;
	fortress->vmax         = ( FORTRESS_VELOCITY / entity_get_fps_ratio() );
	fortress->vx           = ( FORTRESS_VELOCITY / entity_get_fps_ratio() ) * cos( angle );
	fortress->vy           = ( FORTRESS_VELOCITY / entity_get_fps_ratio() ) * sin( angle );
	fortress->check_func   = fortress_check;
	fortress->update_func  = fortress_update;
	fortress->draw_func    = fortress_draw;
}


/*!
 * Computes new position just for collision check.
 * Note that this function does not compute real position.
 * After this, execute entity_update() to compute real position.
 * This function returns 0 when no collision check is needed 
 * because enemy is not alive ...
 *
 * No null check is executed because this is a back-end of entity_check().
 */
int
fortress_check( Entity* fortress, RLrect* rect, double* vx, double* vy )
{
	if ( ! entity_is_alive( fortress ) ) {
		return 0;
	}

	rect->lefttop_x = fortress->px + fortress->vx - ( fortress->width  * 0.45 );
	rect->lefttop_y = fortress->py + fortress->vy - ( fortress->height * 0.45 );
	rect->width     = ( int )( fortress->width  * 0.9 );
	rect->height    = ( int )( fortress->height * 0.9 );

	(*vx) = fortress->vx;
	(*vy) = fortress->vy;

	return 1;
}


/*!
 * Commits new position, status, etc. at the next frame.
 * No null check is executed because this is a back-end of entity_update().
 * spaceship can be NULL, which will be ignored in this function.
 */
void
fortress_update( Entity* fortress, Entity* spaceship )
{
	if ( fortress->is_exploding ) {
		if ( spaceship != NULL ) {
			explosion_update( &( fortress->explosion ), spaceship->px, spaceship->py );
		}
		else {
			explosion_update( &( fortress->explosion ), -1, -1 );
		}

		if ( ! explosion_is_exploding( &( fortress->explosion ) ) ) {
			fortress->is_alive     = 0;
			fortress->is_exploding = 0;
		}
	}
	else if ( fortress->is_alive ) {
		fortress->effect_animation_frame_count ++;

		if ( fortress->frames_to_next_launch > 0 ) {
			fortress->frames_to_next_launch --;
		}

		if ( spaceship != NULL ) {
			fortress_launch_missile( fortress, spaceship );
		}

		fortress->px += fortress->vx;
		fortress->py += fortress->vy;
	}
}


static void
fortress_launch_missile( Entity* fortress, Entity* spaceship )
{
	if ( spaceship == NULL ) {
		return;
	}

	{
		double mx = fortress->px;
		double my = fortress->py;
		double px;
		double py;
		double distance;

		entity_get_position( spaceship, &px, &py );
		distance = DISTANCE( fortress->px, fortress->py, px, py );

		if ( distance < FORTRESS_LAUNCH_DISTANCE_S && 
		     fortress->frames_to_next_launch <= 0 ) {

			rlib_sound_play( spaceship->sound, 
					 spaceship->wave_index[ INDEX_WAVE_MISSILE ], 0 );

			fortress->frames_to_next_launch = 
				( FRAME_COUNT_TO_NEXT_LAUNCH * entity_get_fps_ratio() );

			missile_launch_blast( fortress->gt_app, fortress->missiles, 
				      MISSILE_FIREBOMB, &( fortress->screen ), mx, my );
		}
		else if ( distance < FORTRESS_LAUNCH_DISTANCE_L && 
			  fortress->frames_to_next_launch <= 0 ) {

			rlib_sound_play( spaceship->sound, 
					 spaceship->wave_index[ INDEX_WAVE_DOUGHNUT ], 0 );

			fortress->frames_to_next_launch = 
				( FRAME_COUNT_TO_NEXT_LAUNCH * entity_get_fps_ratio() );

			missile_launch_single_shot( fortress->gt_app, fortress->missiles, 
					    MISSILE_DOUGHNUT, &( fortress->screen ), 
					    mx, my, px - mx, py - my );
		}
	}
}


/*!
 * Draws an enemy (fortress) onto the screen.
 * No null check is executed because this is a back-end of entity_draw().
 */
void
fortress_draw( Entity* fortress )
{
	if ( entity_is_exploding( fortress ) ) {
		explosion_draw( &( fortress->explosion ) );
		return;
	}

	if ( fortress->is_alive ) {
		fortress_draw_body( fortress );
	}
}


static void
fortress_draw_body( Entity* fortress )
{
#define FORTRESS_BODY_POINTS_NUM 37

	const double points[ FORTRESS_BODY_POINTS_NUM ][2] = {
		{   24.0,   60.0 },
		{   24.0,   48.0 },
		{   50.0,   48.0 },
		{   50.0,   36.0 },
		{   76.0,   36.0 },
		{   76.0,   24.0 },
		{  112.0,   24.0 },
		{  112.0,   12.0 },
		{  150.0,   12.0 },
		{  150.0, - 12.0 },
		{  112.0, - 12.0 },
		{  112.0, - 24.0 },
		{   76.0, - 24.0 },
		{   76.0, - 36.0 },
		{   50.0, - 36.0 },
		{   50.0, - 48.0 },
		{   24.0, - 48.0 },
		{   24.0, - 60.0 },
		{ - 24.0, - 60.0 },
		{ - 24.0, - 48.0 },
		{ - 50.0, - 48.0 },
		{ - 50.0, - 36.0 },
		{ - 76.0, - 36.0 },
		{ - 76.0, - 24.0 },
		{ -112.0, - 24.0 },
		{ -112.0, - 12.0 },
		{ -150.0, - 12.0 },
		{ -150.0,   12.0 },
		{ -112.0,   12.0 },
		{ -112.0,   24.0 },
		{ - 76.0,   24.0 },
		{ - 76.0,   36.0 },
		{ - 50.0,   36.0 },
		{ - 50.0,   48.0 },
		{ - 24.0,   48.0 },
		{ - 24.0,   60.0 },
		{   24.0,   60.0 }
	};

	Point converted[ FORTRESS_BODY_POINTS_NUM ];
	int screen_x;
	int screen_y;
	int culling = 1;
	int i;

	SCREEN_POSITION( fortress, screen_x, screen_y );

	for ( i = 0; i < FORTRESS_BODY_POINTS_NUM; i ++ ) {
		converted[i].x = ( short )( screen_x + points[i][0] );
		converted[i].y = ( short )( screen_y + points[i][1] );

		if ( rlib_rect_within( &( fortress->screen ), converted[i].x, converted[i].y ) ) {
			culling = 0;
		}
	}

	if ( culling ) {
		return;
	}
	
	rlib_gt_set_foreground_color( fortress->gt_app, RL_GT_MAX_RGB, RL_GT_MAX_RGB, 0 );
	rlib_gt_draw_lines( fortress->gt_app, converted, FORTRESS_BODY_POINTS_NUM );

	if ( fortress->effect_animation_frame_count %
	     ( int )entity_get_framecount_effect_animation() > 
	     entity_get_framecount_effect_animation() / 2 ) {

		rlib_gt_set_foreground_color( fortress->gt_app, RL_GT_MAX_RGB, 0, 0 );
		fortress_draw_circle( fortress, screen_x, screen_y, 0 );
		rlib_gt_set_foreground_color( fortress->gt_app, 0, 0, RL_GT_MAX_RGB );
		fortress_draw_circle( fortress, screen_x, screen_y, 1 );
	}
	else {
		rlib_gt_set_foreground_color( fortress->gt_app, RL_GT_MAX_RGB, 0, 0 );
		fortress_draw_circle( fortress, screen_x, screen_y, 1 );
		rlib_gt_set_foreground_color( fortress->gt_app, 0, 0, RL_GT_MAX_RGB );
		fortress_draw_circle( fortress, screen_x, screen_y, 0 );
	}
}


static void
fortress_draw_circle( Entity* fortress, int screen_x, int screen_y, int index_offset )
{
#define NUM_OF_CIRCLE   12
#define DISTANCE_OFFSET 24
#define UNIT_ANGLE_DEG  15
#define DIAMETER         5

	double angle;
	int circle_x;
	int circle_y;
	int i;

	for ( i = 0; i < NUM_OF_CIRCLE; i ++ ) {
		angle = ( i * 2 + index_offset ) * UNIT_ANGLE_DEG / ( 2 * PI );
		circle_x = screen_x - DISTANCE_OFFSET * sin( angle );
		circle_y = screen_y + DISTANCE_OFFSET * cos( angle ); 
		rlib_gt_fill_arc( fortress->gt_app, circle_x, circle_y, 
				  DIAMETER, DIAMETER, 0, 360 * 64 );
	}
}


/* End of fortress.c */
