/*
 * Copyright (c) 2005-2007 Ryo Ichinose
 *
 * This file is part of Space Maze.
 * 
 * Space Maze is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; version 2 of the License.
 * 
 * Space Maze is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with Space Maze; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

/*!
 * @file
 * @brief Handles enemy (missiler) and draws it onto the game screen.
 *
 * "$Id: missiler.c,v 1.2 2007/11/04 11:41:23 ryo Exp $"
 */

#include <math.h>
#include <stdio.h>
#include <stdlib.h>

#include "RLgt.h"
#include "RLrect.h"
#include "RLsound.h"
#include "enemy.h"
#include "entity.h"
#include "explosion.h"
#include "missile.h"
#include "spaceship.h"
#include "resource.h"


#define FRAME_COUNT_TO_NEXT_LAUNCH  180

#define MISSILER_WIDTH              120
#define MISSILER_HEIGHT             40
#define MISSILER_LAUNCH_DISTANCE    250
#define MISSILER_MAX_ENERGY         250
#define MISSILER_STRENGTH           10
#define MISSILER_VELOCITY           1.5


static void missiler_draw_body( Entity* missiler );
static void missiler_launch_missile( Entity* missiler, Entity* spaceship );


/*!
 * Initializes enemy (missiler) specific data.
 * No null check is executed because this is a back-end of enemy_init().
 */
void
missiler_init( Entity* missiler )
{
	missiler->width             = MISSILER_WIDTH;
	missiler->height            = MISSILER_HEIGHT;
	missiler->energy            = MISSILER_MAX_ENERGY;
	missiler->strength          = MISSILER_STRENGTH;
	missiler->vmax              = ( MISSILER_VELOCITY / entity_get_fps_ratio() );
	missiler->vx                = ( MISSILER_VELOCITY / entity_get_fps_ratio() );
	missiler->vy                = 0;
	missiler->set_velocity_func = missiler_set_velocity;
	missiler->check_func        = missiler_check;
	missiler->update_func       = missiler_update;
	missiler->draw_func         = missiler_draw;
}


/*!
 * Sets the velocity of an enemy (missiler) correspond 
 * to the world coordinates.
 */
void
missiler_set_velocity( Entity* missiler, double vx, double vy )
{
	if ( missiler == NULL ) {
		return;
	}

	missiler->vx = vx;
	/* Ignores y component. */
}


/*!
 * Computes new position just for collision check.
 * Note that this function does not compute real position.
 * After this, execute entity_update() to compute real position.
 * This function returns 0 when no collision check is needed 
 * because enemy is not alive ...
 *
 * No null check is executed because this is a back-end of entity_check().
 */
int
missiler_check( Entity* missiler, RLrect* rect, double* vx, double* vy )
{
	if ( ! entity_is_alive( missiler ) ) {
		return 0;
	}

	rect->lefttop_x = missiler->px + missiler->vx - missiler->width  / 2.0;
	rect->lefttop_y = missiler->py + missiler->vy - missiler->height / 2.0;
	rect->width     = missiler->width;
	rect->height    = missiler->height;

	(*vx) = missiler->vx;
	(*vy) = missiler->vy;

	return 1;
}


/*!
 * Commits new position, status, etc. at the next frame.
 * No null check is executed because this is a back-end of entity_update().
 * spaceship can be NULL, which will be ignored in this function.
 */
void
missiler_update( Entity* missiler, Entity* spaceship )
{
	if ( missiler->is_exploding ) {
		if ( spaceship != NULL ) {
			explosion_update( &( missiler->explosion ), spaceship->px, spaceship->py );
		}
		else {
			explosion_update( &( missiler->explosion ), -1, -1 );
		}

		if ( ! explosion_is_exploding( &( missiler->explosion ) ) ) {
			missiler->is_alive     = 0;
			missiler->is_exploding = 0;
		}
	}
	else if ( missiler->is_alive ) {
		if ( missiler->frames_to_next_launch > 0 ) {
			missiler->frames_to_next_launch --;
		}

		if ( spaceship != NULL ) {
			missiler_launch_missile( missiler, spaceship );
		}

		missiler->px += missiler->vx;
		missiler->py += missiler->vy;
	}
}


static void
missiler_launch_missile( Entity* missiler, Entity* spaceship )
{
	if ( spaceship == NULL ) {
		return;
	}

	{
		double mx = missiler->px;
		double my = missiler->py;
		double px;
		double py;
		double distance;

		entity_get_position( spaceship, &px, &py );
		distance = DISTANCE( missiler->px, missiler->py, px, py );

		if ( distance < MISSILER_LAUNCH_DISTANCE && 
		     missiler->frames_to_next_launch <= 0 ) {

			rlib_sound_play( spaceship->sound, 
					 spaceship->wave_index[ INDEX_WAVE_MISSILE ], 0 );

			missiler->frames_to_next_launch = 
				( FRAME_COUNT_TO_NEXT_LAUNCH * entity_get_fps_ratio() );

			missile_launch_blast( missiler->gt_app, missiler->missiles, 
				      MISSILE_CANNON, &( missiler->screen ), mx, my );
		}
	}
}


/*!
 * Draws an enemy (missiler) onto the screen.
 * No null check is executed because this is a back-end of entity_draw().
 */
void
missiler_draw( Entity* missiler )
{
	if ( entity_is_exploding( missiler ) ) {
		explosion_draw( &( missiler->explosion ) );
		return;
	}

	if ( missiler->is_alive ) {
		missiler_draw_body( missiler );
	}
}


static void
missiler_draw_body( Entity* missiler )
{
#define MISSILER_BODY_POINTS_NUM 21

	const double points[ MISSILER_BODY_POINTS_NUM ][2] = {
		{  10.0,  20.0 },
		{  25.0,   5.0 },
		{  35.0,   5.0 },
		{  35.0,  20.0 },
		{  40.0,  20.0 },
		{  40.0, -20.0 },
		{  35.0, -20.0 },
		{  35.0, - 5.0 },
		{  25.0, - 5.0 },
		{  10.0, -20.0 },
		{ -10.0, -20.0 },
		{ -25.0, - 5.0 },
		{ -35.0, - 5.0 },
		{ -35.0, -20.0 },
		{ -40.0, -20.0 },
		{ -40.0,  20.0 },
		{ -35.0,  20.0 },
		{ -35.0,   5.0 },
		{ -25.0,   5.0 },
		{ -10.0,  20.0 },
		{  10.0,  20.0 }
	};

	Point converted[ MISSILER_BODY_POINTS_NUM ];
	int screen_x;
	int screen_y;
	int culling = 1;
	int i;

	SCREEN_POSITION( missiler, screen_x, screen_y );

	for ( i = 0; i < MISSILER_BODY_POINTS_NUM; i ++ ) {
		converted[i].x = ( short )( screen_x + points[i][0] );
		converted[i].y = ( short )( screen_y + points[i][1] );

		if ( rlib_rect_within( &( missiler->screen ), converted[i].x, converted[i].y ) ) {
			culling = 0;
		}
	}

	if ( culling ) {
		return;
	}
	
	rlib_gt_set_foreground_color( missiler->gt_app, RL_GT_MAX_RGB * 0.8, 0, 0 );
	rlib_gt_draw_lines( missiler->gt_app, converted, MISSILER_BODY_POINTS_NUM );
	rlib_gt_set_foreground_color( missiler->gt_app, 0, 0, RL_GT_MAX_RGB );
	rlib_gt_draw_rectangle(
		missiler->gt_app,
		screen_x - MISSILER_WIDTH  / 12.0,
		screen_y - MISSILER_HEIGHT / 8.0,
		MISSILER_WIDTH  / 6.0,
		MISSILER_HEIGHT / 4.0
		);
}


/* End of missiler.c */
