/*
 * Copyright (c) 2005,2006 Ryo Ichinose
 *
 * This file is part of Space Maze.
 * 
 * Space Maze is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; version 2 of the License.
 * 
 * Space Maze is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with Space Maze; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

/*!
 * @file
 * @brief Handles enemy (scull) data and draws it onto the screen.
 *
 * "$Id: scull.c,v 1.1.1.1 2006/07/22 18:07:56 ryo Exp $"
 */

#include <math.h>
#include <stdio.h>
#include <stdlib.h>

#include "RLgt.h"
#include "RLrect.h"
#include "enemy.h"
#include "entity.h"
#include "explosion.h"
#include "missile.h"
#include "spaceship.h"


#define FRAME_COUNT_TO_NEXT_LAUNCH 100

#define SCULL_WIDTH                40
#define SCULL_HEIGHT               40
#define SCULL_LAUNCH_DISTANCE      150
#define SCULL_MAX_ENERGY           150
#define SCULL_STRENGTH             15
#define SCULL_VELOCITY             2.3


static void scull_draw_body( Entity* scull );
static void scull_launch_missile( Entity* scull, Entity* spaceship );


/*!
 * Initializes enemy (scull) specific data.
 * No null check is executed because this is a back-end of enemy_init().
 */
void
scull_init( Entity* scull )
{
	double angle = RANDOM( 2.0 * PI );

	scull->width        = SCULL_WIDTH;
	scull->height       = SCULL_HEIGHT;
	scull->energy       = SCULL_MAX_ENERGY;
	scull->strength     = SCULL_STRENGTH;
	scull->vmax         = ( SCULL_VELOCITY / entity_get_fps_ratio() );
	scull->vx           = ( SCULL_VELOCITY / entity_get_fps_ratio() ) * cos( angle );
	scull->vy           = ( SCULL_VELOCITY / entity_get_fps_ratio() ) * sin( angle );
	scull->check_func   = scull_check;
	scull->update_func  = scull_update;
	scull->draw_func    = scull_draw;
}


/*!
 * Computes new position just for collision check.
 * Note that this function does not compute real position.
 * After this, execute entity_update() to compute real position.
 * This function returns 0 when no collision check is needed 
 * because enemy is not alive ...
 *
 * No null check is executed because this is a back-end of scullheck().
 */
int
scull_check( Entity* scull, RLrect* rect, double* vx, double* vy )
{
	if ( ! entity_is_alive( scull ) ) {
		return 0;
	}

	rect->lefttop_x = scull->px + scull->vx - ( scull->width  * 0.45 );
	rect->lefttop_y = scull->py + scull->vy - ( scull->height * 0.45 );
	rect->width     = ( int )scull->width;
	rect->height    = ( int )scull->height;

	(*vx) = scull->vx;
	(*vy) = scull->vy;

	return 1;
}


/*!
 * Commits new position, status, etc. at the next frame.
 * No null check is executed because this is a back-end of entity_update().
 * spaceship can be NULL, which will be ignored in this function.
 */
void
scull_update( Entity* scull, Entity* spaceship )
{
	if ( scull->is_exploding ) {
		if ( spaceship != NULL ) {
			explosion_update( &( scull->explosion ), spaceship->px, spaceship->py );
		}
		else {
			explosion_update( &( scull->explosion ), -1, -1 );
		}

		if ( ! explosion_is_exploding( &( scull->explosion ) ) ) {
			scull->is_alive     = 0;
			scull->is_exploding = 0;
		}
	}
	else if ( scull->is_alive ) {
		if ( scull->frames_to_next_launch > 0 ) {
			scull->frames_to_next_launch --;
		}

		if ( spaceship != NULL ) {
			scull_launch_missile( scull, spaceship );
		}

		scull->px += scull->vx;
		scull->py += scull->vy;
	}
}


static void
scull_launch_missile( Entity* scull, Entity* spaceship )
{
	if ( spaceship == NULL ) {
		return;
	}

	{
		double mx = scull->px;
		double my = scull->py;
		double px;
		double py;
		double distance;

		entity_get_position( spaceship, &px, &py );
		distance = DISTANCE( scull->px, scull->py, px, py );

		if ( distance < SCULL_LAUNCH_DISTANCE && scull->frames_to_next_launch <= 0 ) {
			scull->frames_to_next_launch = 
				( FRAME_COUNT_TO_NEXT_LAUNCH * entity_get_fps_ratio() );

			missile_launch_three_way( scull->gt_app, scull->missiles, MISSILE_RIPPLE,
				    &( scull->screen ), mx, my, px - mx, py - my );
		}
	}
}


/*!
 * Draws an enemy (scull) onto the screen.
 * No null check is executed because this is a back-end of entity_draw().
 */
void
scull_draw( Entity* scull )
{
	if ( entity_is_exploding( scull ) ) {
		explosion_draw( &( scull->explosion ) );
		return;
	}

	if ( scull->is_alive ) {
		scull_draw_body( scull );
	}
}


static void
scull_draw_body( Entity* scull )
{
#define SCULL_BODY_POINTS_NUM 11

	const double points[ SCULL_BODY_POINTS_NUM ][2] = {
		{  15.0, -20.0 },
		{  20.0, -15.0 },
		{  20.0,   5.0 },
		{  15.0,  10.0 },
		{  15.0,  20.0 },
		{ -15.0,  20.0 },
		{ -15.0,  10.0 },
		{ -20.0,   5.0 },
		{ -20.0, -15.0 },
		{ -15.0, -20.0 },
		{  15.0, -20.0 }
	};

	Point converted[ SCULL_BODY_POINTS_NUM ];
	int screen_x;
	int screen_y;
	int culling = 1;
	int i;

	SCREEN_POSITION( scull, screen_x, screen_y );

	for ( i = 0; i < SCULL_BODY_POINTS_NUM; i ++ ) {
		converted[i].x = ( short )( screen_x + points[i][0] );
		converted[i].y = ( short )( screen_y + points[i][1] );

		if ( rlib_rect_within( &( scull->screen ), converted[i].x, converted[i].y ) ) {
			culling = 0;
		}
	}

	if ( culling ) {
		return;
	}

	rlib_gt_set_foreground_color( scull->gt_app, RL_GT_MAX_RGB, RL_GT_MAX_RGB, RL_GT_MAX_RGB );
	rlib_gt_draw_lines( scull->gt_app, converted, SCULL_BODY_POINTS_NUM );
	rlib_gt_set_foreground_color( scull->gt_app, RL_GT_MAX_RGB, 0, 0 );

	rlib_gt_draw_arc( 
		scull->gt_app,
		screen_x - SCULL_WIDTH  * 3.0 / 8.0, 
		screen_y - SCULL_HEIGHT * 3.0 / 8.0,
		SCULL_WIDTH  / 4.0,
		SCULL_HEIGHT / 4.0,
		0,
		360 * 64
		);

	rlib_gt_draw_arc( 
		scull->gt_app,
		screen_x + SCULL_WIDTH  / 8.0,
		screen_y - SCULL_HEIGHT * 3.0 / 8.0,
		SCULL_WIDTH  / 4.0,
		SCULL_HEIGHT / 4.0,
		0,
		360 * 64
		);

	rlib_gt_draw_rectangle(
		scull->gt_app,
		screen_x - SCULL_WIDTH  / 16.0,
		screen_y - SCULL_HEIGHT / 16.0,
		SCULL_WIDTH  / 8.0,
		SCULL_HEIGHT / 4.0
		);

	rlib_gt_draw_rectangle(
		scull->gt_app,
		screen_x - SCULL_WIDTH  / 4.0,
		screen_y + SCULL_HEIGHT / 4.0,
		SCULL_WIDTH  / 2.0,
		SCULL_HEIGHT / 5.0
		);
}


/* End of scull.c */
