/*
 * Copyright (c) 2005-2007 Ryo Ichinose
 *
 * This file is part of Space Maze.
 * 
 * Space Maze is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; version 2 of the License.
 * 
 * Space Maze is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with Space Maze; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

/*!
 * @file
 * Handles entity data and draws it onto the screen.
 * This file is the basic component of various objects.
 *
 * "$Id: entity.c,v 1.2 2007/11/04 11:41:23 ryo Exp $"
 */

#include <math.h>
#include <string.h>

#include "RLgt.h"
#include "RLrect.h"
#include "RLsound.h"
#include "entity.h"
#include "explosion.h"
#include "resource.h"


#define ENTITY_FRAME_COUNT_EFFECT_BLINK     45
#define ENTITY_FRAME_COUNT_EFFECT_ANIMATION 30


static int* f_image_index;
static int* f_wave_index;
static double f_ratio;


/*!
 * Sets fps ratio ( current fps / ( default fps = 60 ) ).
 */
void 
entity_set_fps_ratio( double ratio )
{
	f_ratio = ratio;
}


/*!
 * Gets fps ratio ( current fps / ( default fps = 60 ) ).
 */
double
entity_get_fps_ratio()
{
	return f_ratio;
}


int  
entity_get_framecount_effect_blink()
{
	return ( int )( ENTITY_FRAME_COUNT_EFFECT_BLINK * f_ratio );
}


int  
entity_get_framecount_effect_animation()
{
	return ( int )( ENTITY_FRAME_COUNT_EFFECT_ANIMATION * f_ratio );
}


/*!
 * Sets image and wave data indexes to refer to the real data
 * stored in GUI toolkit context and sound library context.
 *
 * @param image_index [in] Image indexes.
 * @param wave_index  [in] Wave data indexes.
 */
void
entity_set_resource_indexes( int* image_index, int* wave_index )
{
	f_image_index = image_index;
	f_wave_index  = wave_index;
}


/*!
 * Initializes entity data structure.
 */
void
entity_init( Entity* entity, RLgtApp* app, RLsoundContext* sound, RLrect* screen )
{
	/*
	 * "sound" can be NULL when the entity doesn't use sound.
	 */
	if ( entity == NULL || app == NULL || screen == NULL ) {
		return;
	}

	memset( entity, 0, sizeof( Entity ) );

	entity->gt_app          = app;
	entity->sound           = sound;
	entity->image_index     = f_image_index;
	entity->wave_index      = f_wave_index;
	entity->is_alive        = 1;
	entity->vmax            = 0.0;
	entity->screen_center_x = 0.0;
	entity->screen_center_y = 0.0;
	entity->initial_px      = 0.0;
	entity->initial_py      = 0.0;
	entity->px              = 0.0;
	entity->py              = 0.0;
	entity->vx              = 0.0;
	entity->vy              = 0.0;
	entity->ax              = 0.0;
	entity->ay              = 0.0;
	entity->prot            = 0.0;
	entity->vrot            = 0.0;
	entity->arot            = 0.0;

	if ( screen != NULL ) {
		entity->screen.lefttop_x = screen->lefttop_x;
		entity->screen.lefttop_y = screen->lefttop_y;
		entity->screen.width     = screen->width;
		entity->screen.height    = screen->height;
	}
}


/*!
 * Checks if specified entity is alive or not.
 * Returns 1 when alive, 0 when not or entity is NULL.
 */
int
entity_is_alive( Entity* entity )
{
	if ( entity == NULL ) {
		return 0;
	}

	return entity->is_alive;
}


/*!
 * Checks if specified entity is exploding or not.
 * Returns 1 when alive, 0 when not or entity is NULL.
 */
int
entity_is_exploding( Entity* entity )
{
	if ( entity == NULL ) {
		return 0;
	}

	return entity->is_exploding;
}


/*!
 * Sets entity status to exploding.
 */	
void
entity_explode( Entity* entity )
{
	if ( entity == NULL ) {
		return;
	}

	if ( entity->explode_func == NULL ) {
		double px;
		double py;

		entity->is_alive     = 0;
		entity->is_exploding = 1;
		entity->energy       = 0;

		rlib_sound_play( entity->sound, entity->wave_index[ INDEX_WAVE_EXPLOSION ], 0 );
		explosion_init( &( entity->explosion ), entity->gt_app, &( entity->screen ), f_image_index );
		entity_get_position( entity, &px, &py );
		explosion_set_position( &( entity->explosion ), px, py );
		return;
	}

	entity->explode_func( entity );
}


/*!
 * Gets strength of the specified entity.
 */
int
entity_get_strength( Entity* entity )
{
	if ( entity == NULL ) {
		return 0;
	}

	return entity->strength;
}


/*!
 * Gets current energy amount of the specified entity.
 */
int
entity_get_energy( Entity* entity )
{
	if ( entity == NULL ) {
		return 0;
	}

	return entity->energy;
}


/*!
 * Decreases entity energy with specified amount.
 * When the amount is under 0, it'll be ignored.
 * If entity energy reaches 0 by this decreasing,
 * entity_explode() will be called.
 */
void
entity_decrease_energy( Entity* entity, int amount )
{
	if ( entity == NULL || amount < 0 ) {
		return;
	}

	if ( entity->decrease_func == NULL ) {
		entity->energy -= amount;
		entity->effect_blink_frame_count = entity_get_framecount_effect_blink();
	}
	else {
		entity->decrease_func( entity, amount );
	}

	if ( entity->energy <= 0 ) {
		entity_explode( entity );
	}
}


/*!
 * Make specified entity blink.
 */
void
entity_blink( Entity* entity )
{
	if ( entity == NULL ) {
		return;
	}

	if ( entity->blink_func == NULL ) {
		entity->effect_blink_frame_count = entity_get_framecount_effect_blink();
	}
	else {
		entity->blink_func( entity );
	}
}


/*!
 * Binds the specified RLlist to the missile list of the entity.
 */
void
entity_set_missile_list( Entity* entity, RLlist* missiles )
{
	if ( entity == NULL || missiles == NULL ) {
		return;
	}

	entity->missiles = missiles;
}


/*!
 * Gets the width of an entity.
 */
int
entity_get_width( Entity* entity )
{
	if ( entity == NULL ) {
		return 0;
	}

	return entity->width;
}


/*!
 * Gets the height of an entity.
 */
int
entity_get_height( Entity* entity )
{
	if ( entity == NULL ) {
		return 0;
	}

	return entity->height;
}


/*!
 * Gets the position of an entity correspond to the world coordinates.
 */
void
entity_get_position( Entity* entity, double* x, double * y )
{
	if ( entity == NULL || x == NULL || y == NULL ) {
		return;
	}

	*x = entity->px;
	*y = entity->py;
}


/*!
 * Sets the position of an entity correspond to the world coordinates.
 */
void
entity_set_position( Entity* entity, double x, double y )
{
	if ( entity == NULL ) {
		return;
	}

	entity->px = x;
	entity->py = y;
}


/*!
 * Gets the velocity of an entity correspond the world coordinates.
 */
void
entity_get_velocity( Entity* entity, double* vx, double* vy )
{
	if ( entity == NULL || vx == NULL || vy == NULL ) {
		return;
	}

	*vx = entity->vx;
	*vy = entity->vy;
}


/*!
 * Sets the velocity of an entity correspond to the world coordinates.
 */
void
entity_set_velocity( Entity* entity, double vx, double vy )
{
	if ( entity == NULL ) {
		return;
	}

	if ( entity->set_velocity_func == NULL ) {
		entity->vx = vx;
		entity->vy = vy;
		return;
	}

	entity->set_velocity_func( entity, vx, vy );
}


/*!
 * Gets the rotation angle of an entity correspond the world coordinates.
 */
double 
entity_get_rotation( Entity* entity )
{
	if ( entity == NULL ) {
		return 0.0;
	}

	return entity->prot;
}


/*!
 * Sets the rotation angle of an entity correspond to the world coordinates.
 */
void
entity_set_rotation( Entity* entity, double prot )
{
	if ( entity == NULL ) {
		return;
	}

	entity->prot = prot;
}


/*!
 * Checks if the specified entity is a key guardian and the key is released.
 * Returns 1 when the key is released and the function call is the first time.
 * Otherwise returns 0.
 */
int
entity_key_released( Entity* entity )
{
	if ( entity == NULL ) {
		return 0;
	}

	if ( entity->key_released_func == NULL ) {
		return 0;
	}

	return entity->key_released_func( entity );
}


/*!
 * Checks input and computes new position just for collision check.
 * Note that this function does not compute real position.
 * After this, execute entity_update() to commit new position.
 * This function returns 0 when no collision check is needed because
 * an entity is not alive.
 *
 * @param entity [i/o] An entity data storage.
 * @param rect   [out] Rectangle that represents new position.
 * @param vx     [out] X value of new velocity.
 * @param vy     [out] Y value of new velocity.
 */
int
entity_check( Entity* entity, RLrect* rect, double* vx, double* vy )
{
	if ( entity == NULL || entity->gt_app == NULL || 
	     rect == NULL || vx == NULL || vy == NULL ) {
		return 0;
	}

	if ( ! entity_is_alive( entity ) ) {
		return 0;
	}

	if ( entity->check_func == NULL ) {
		rect->lefttop_x = entity->px + entity->vx - entity->width;
		rect->lefttop_y = entity->py + entity->vy - entity->height;
		rect->width     = entity->width;
		rect->height    = entity->height;

		(*vx) = entity->vx;
		(*vy) = entity->vy;

		return 1;
	}

	return entity->check_func( entity, rect, vx, vy );
}


/*!
 * Commits new position, status, etc. at the next frame.
 */
void
entity_update( Entity* entity, Entity* spaceship )
{
	if ( entity == NULL || entity->gt_app == NULL ) {
		return;
	}

	if ( entity->update_func == NULL ) {
		entity->vx += entity->ax;
		entity->vy += entity->ay;
		entity->px += entity->vx;
		entity->py += entity->vy;
	}
	else {
		entity->update_func( entity, spaceship );
	}

	if ( spaceship != NULL ) {
		entity_get_position( spaceship, 
		     &( entity->screen_center_x ), &( entity->screen_center_y ) );
	}
}


/*!
 * Draws an entity onto the game screen.
 */
void
entity_draw( Entity* entity )
{
	if ( entity == NULL || entity->gt_app == NULL || entity->draw_func == NULL ) {
		return;
	}

	entity->draw_func( entity );
}


/*!
 * First, calculates to-be position of each entity at the next frame.
 * Then checks if each entity collides each other, and changes velocity 
 * of both entity if it's necessary to realize reflection. Returns 1 
 * when changing velocity is executed, 0 when no change is applied.
 */
int
entity_calculate_reflection( Entity* entity1, Entity* entity2 )
{
	RLrect rect1;
	RLrect rect2;
	double distance;
	double rx12;
	double ry12;
	double velocity1;
	double velocity2;
	double vx1;
	double vy1;
	double vx2;
	double vy2;

	if ( entity1 == NULL || entity1->gt_app == NULL || 
	     entity2 == NULL || entity2->gt_app == NULL ) {
		return 0;
	}

	if ( ! entity_check( entity1, &rect1, &vx1, &vy1 ) ) {
		return 0;
	}

	if ( ! entity_check( entity2, &rect2, &vx2, &vy2 ) ) {
		return 0;
	}

	if ( ! rlib_rect_overlap( &rect1, &rect2 ) ) {
		return 0;
	}

	/*
	 * Calculates a vector to connect each other's center point, 
	 * which direction is from entity1 to entity2.
	 * Suppose that a reaction occures along this line.
	 */
	rx12 = ( rect2.lefttop_x + rect2.width  / 2 ) - ( rect1.lefttop_x + rect1.width  / 2 );
	ry12 = ( rect2.lefttop_y + rect2.height / 2 ) - ( rect1.lefttop_y + rect1.height / 2 );

	distance = sqrt( pow( rx12, 2 ) + pow( ry12, 2 ) );

	if ( distance >= THRESHOLD ) { /* To avoid 0 division. */
		rx12 /= distance;
		ry12 /= distance;
	}
	else {
		rx12 = ry12 = 0.0;
	}

	/*
	 * Calculates new velocity by adding each other's velocity on the 
	 * same direction of the vector that is already computed above.
	 */
	velocity1 = sqrt( pow( vx1, 2 ) + pow( vy1, 2 ) );
	velocity2 = sqrt( pow( vx2, 2 ) + pow( vy2, 2 ) );

	vx1 += velocity2 * ( -1.0 ) * rx12;
	vy1 += velocity2 * ( -1.0 ) * ry12;

	vx2 += velocity1 * rx12;
	vy2 += velocity1 * ry12;

	/*
	 * Then normalizes each velocity to be its maximum value, 
	 * and commits new velocity at the next frame.
	 */
	velocity1 = sqrt( pow( vx1, 2 ) + pow( vy1, 2 ) );
	velocity2 = sqrt( pow( vx2, 2 ) + pow( vy2, 2 ) );

	vx1 = vx1 * ( entity1->vmax / velocity1 );
	vy1 = vy1 * ( entity1->vmax / velocity1 );

	vx2 = vx2 * ( entity2->vmax / velocity2 );
	vy2 = vy2 * ( entity2->vmax / velocity2 );

	entity_set_velocity( entity1, vx1, vy1 );
	entity_set_velocity( entity2, vx2, vy2 );

	return 1;
}


/* End of entity.c */
