/*
 * Copyright (c) 2006 Ryo Ichinose
 *
 * This file is part of Space Maze.
 * 
 * Space Maze is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; version 2 of the License.
 * 
 * Space Maze is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with Space Maze; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

/*!
 * @file
 * @brief Map data utilities.
 *
 * "$Id: map_util.c,v 1.1.1.1 2006/07/22 18:07:56 ryo Exp $"
 */

#ifdef HAVE_CONFIG_H
#  include "config.h"
#endif

#include <stdio.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <fcntl.h>

#include "RLlogger.h"
#include "map_util.h"


static unsigned char map_checksum( unsigned int value );


/*!
 * Checks the format of the specified map file and then loads file image 
 * into the specified buffer. If the specified map file can be used, 
 * returns 0. Returns -1 if the specified map file is invalid format.
 */
int
map_check_and_load_datafile( char* file_name, unsigned char* buf )
{
	char logbuf[1024];
	struct stat file_stat;
	int fd = -1;

	if ( ( fd = open( file_name, O_RDONLY ) ) == -1 ) {
		snprintf( logbuf, sizeof( logbuf ), "Failed to open map data : %s", file_name );
		rlib_logger_fatal( logbuf );
		goto err_map_check_and_load_datafile;		
	}

	if ( fstat( fd, &file_stat ) == -1 ) {
		snprintf( logbuf, sizeof( logbuf ), "fstat() failed to access map data info : %s", file_name );
		rlib_logger_fatal( logbuf );
		goto err_map_check_and_load_datafile;
	}

	if ( file_stat.st_size != MAP_FILE_SIZE ) {
		snprintf( logbuf, sizeof( logbuf ), "Map data file size is invalid : %s", file_name );
		rlib_logger_fatal( logbuf );
		goto err_map_check_and_load_datafile;
	}

	if ( read( fd, buf, MAP_FILE_SIZE ) != MAP_FILE_SIZE ) {
		snprintf( logbuf, sizeof( logbuf ), "read() failed to read map data : %s", file_name );
		rlib_logger_fatal( logbuf );
		goto err_map_check_and_load_datafile;
	}

	if ( memcmp( buf, MAP_FILE_MAGIC, strlen( MAP_FILE_MAGIC ) ) != 0 ) {
		snprintf( logbuf, sizeof( logbuf ), 
			  "Specified file does not contain magic number : %s", file_name  );
		rlib_logger_fatal( logbuf );
		goto err_map_check_and_load_datafile;
	}

	{
		unsigned int value = 0;
		int i;

		for ( i = 0; i < MAP_FILE_SIZE - MAP_CSUM_SIZE; i ++ ) {
			value += buf[i];
		}

		if ( map_checksum( value ) != buf[ MAP_FILE_SIZE - 1 ] ) {
			snprintf( logbuf, sizeof( logbuf ), "Checksum of map data is invalid : %s", file_name );
			rlib_logger_fatal( logbuf );
			goto err_map_check_and_load_datafile;
		}
	}		

	close( fd );
	return 0;

 err_map_check_and_load_datafile:

	if ( fd != -1 ) {
		close( fd );
	}

	return -1;
}


/*!
 * Checks the format of the specified map source file and then 
 * saves it into map data file. If the format of the specified 
 * map source file is invalid, returns -1. 
 * Returns 0 when successfully processed.
 */
int
map_check_and_save_datafile( char* infile, char* outfile )
{
	unsigned char ibuf[ MAP_COL_NUM + 32 ]; /* 32 is just the margin size. */
	unsigned int  value = 0;
	char  logbuf[1024];
	FILE* fp = NULL;
	int   fd = -1;
	int   i;

	if ( infile == NULL || outfile == NULL ) {
		snprintf( logbuf, sizeof( logbuf ), "Input name or output name of map data is NULL." );
		rlib_logger_fatal( logbuf );
		goto err_map_check_and_save_datafile;
	}

	if ( ( fp = fopen( infile, "r" ) ) == NULL ) {
		snprintf( logbuf, sizeof( logbuf ), "Failed to open map source file : %s", infile );
		rlib_logger_fatal( logbuf );
		goto err_map_check_and_save_datafile;
	}

	{
		int flags = O_WRONLY | O_CREAT | O_TRUNC;
		int mode  = S_IRUSR | S_IWUSR | S_IRGRP | S_IROTH;

		if ( ( fd = open( outfile, flags, mode ) ) == -1 ) {
			snprintf( logbuf, sizeof( logbuf ), "Failed to open map data file : %s", outfile );
			rlib_logger_fatal( logbuf );
			goto err_map_check_and_save_datafile;
		}
	}

	if ( write( fd, MAP_FILE_MAGIC, strlen( MAP_FILE_MAGIC ) ) != strlen( MAP_FILE_MAGIC ) ) {
		snprintf( logbuf, sizeof( logbuf ), "Failed to write map data magic number : %s", outfile );
		rlib_logger_fatal( logbuf );
		goto err_map_check_and_save_datafile;
	}

	for ( i = 0; i < strlen( MAP_FILE_MAGIC ); i ++ ) {
		value += MAP_FILE_MAGIC[i];
	}

	{
		int row;
		int col;

		for ( row = 0; row < MAP_ROW_NUM; row ++ ) {
			if ( fgets( ( char* )ibuf, sizeof( ibuf ), fp ) == NULL ) {
				snprintf( logbuf, sizeof( logbuf ), "Failed to read a map source line : %s", infile );
				rlib_logger_fatal( logbuf );
				goto err_map_check_and_save_datafile;
			}

			for ( col = 0; col < MAP_COL_NUM; col ++ ) {
				if ( ibuf[ col ] == '\n' ) {
					snprintf( logbuf, sizeof( logbuf ), 
					  "Map source line must have %d characters : %s", MAP_COL_NUM, infile );
					rlib_logger_fatal( logbuf );
					goto err_map_check_and_save_datafile;
				}
				value += ibuf[ col ];
			}

			if ( write( fd, ibuf, MAP_COL_NUM ) != MAP_COL_NUM ) {
				snprintf( logbuf, sizeof( logbuf ), "Failed to write map data : %s", outfile );
				rlib_logger_fatal( logbuf );
				goto err_map_check_and_save_datafile;
			}
		}
	}

	{
		unsigned char checksum = map_checksum( value );

		if ( write( fd, &checksum, 1 ) != 1 ) {
			snprintf( logbuf, sizeof( logbuf ), "Failed to write map checksum : %s", outfile );
			rlib_logger_fatal( logbuf );
			goto err_map_check_and_save_datafile;
		}
	}

	fclose( fp );
	close( fd );
	return 0;

 err_map_check_and_save_datafile:

	if ( fp != NULL ) {
		fclose( fp );
	}

	if ( fd != -1 ) {
		close( fd );
	}

	return -1;
}


static unsigned char 
map_checksum( unsigned int value )
{
	return ( ~( value & 0xff ) + 1 ) & 0xff;
}


/* End of map_util.c */
