/*
 * Copyright (c) 2005,2006 Ryo Ichinose
 *
 * This file is part of Space Maze.
 * 
 * Space Maze is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; version 2 of the License.
 * 
 * Space Maze is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with Space Maze; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

/*!
 * @file
 * @brief Handles missile entity data and draws it onto the screen.
 *
 * "$Id: missile.c,v 1.1.1.1 2006/07/22 18:07:56 ryo Exp $"
 */

#include <math.h>
#include <stdio.h>
#include <string.h>

#include "RLlist.h"
#include "RLrect.h"
#include "missile.h"
#include "entity.h"


/*!
 * Initializes missile data structures.
 *
 * @param missile [i/o] A missile data storage.
 * @param app     [in]  GUI toolkit context.
 * @param type    [in]  Type of a missile.
 * @param screen  [in]  Screen definition on which missiles should be drawn.
 */
void
missile_init( Entity* missile, RLgtApp* app, MissileType type, RLrect* screen )
{
	if ( missile == NULL || app == NULL || screen == NULL ) {
		return;
	}

	entity_init( missile, app, NULL, screen );
	missile->explode_func = missile_explode;

	switch ( type ) {
		case MISSILE_MINE:
			mine_init( missile );
			break;

		case MISSILE_CANNON:
			cannon_init( missile );
			break;

		case MISSILE_RIPPLE:
			ripple_init( missile );
			break;

		case MISSILE_FIREBOMB:
			firebomb_init( missile );
			break;

		case MISSILE_DOUGHNUT:
			doughnut_init( missile );
			break;

		case MISSILE_LASER:
		default:
			laser_init( missile );
			break;
	}
}


/*!
 * Sets missile status to exploding.
 */
void
missile_explode( Entity* missile )
{
	if ( missile == NULL ) {
		return;
	}

	/*
	 * Currently, I want to make missile disappear right after 
	 * it is set to explode. 
	 * No effect is going to be implemented for a while ...
	 */
	missile->is_alive     = 0;
	missile->is_exploding = 0;
}


/*!
 * Launches a single missile.
 *
 * @param app       [in]  GUI toolkit context.
 * @param missiles  [i/o] Missile list. Launched missile will be inserted to this list.
 * @param type      [in]  Type of a missile.
 * @param screen    [in]  Screen definition.
 * @param launch_x  [in]  X value of launched position.
 * @param launch_y  [in]  Y value of launched position.
 * @param dir_x     [in]  X value of launch direction. Vector (dir_x,dir_y) will be normalized.
 * @param dir_y     [in]  Y value of launch direction. Vector (dir_x,dir_y) will be normalized.
 */
void
missile_launch_single_shot( RLgtApp* app, RLlist* missiles, MissileType type, 
    RLrect* screen, double launch_x, double launch_y, double dir_x, double dir_y )
{
	if ( app == NULL || missiles == NULL || screen == NULL ) {
		return;
	}

	{
		Entity missile;
		double vx;
		double vy;

		NORMALIZE( dir_x, dir_y );

		if ( fabs( dir_x ) < THRESHOLD && fabs( dir_y ) < THRESHOLD ) {
			return;
		}

		missile_init( &missile, app, type, screen );
		vx = missile.vmax * dir_x;
		vy = missile.vmax * dir_y;
		entity_set_position( &missile, launch_x, launch_y );
		entity_set_velocity( &missile, vx, vy );
		entity_set_rotation( &missile, -atan2( vy, vx ) );
		rlib_list_insert( missiles, -1, ( unsigned char*)( &missile ), sizeof( Entity ) );
	}
}


/*!
 * Launches three way missiles.
 *
 * @param app       [in]  GUI toolkit context.
 * @param missiles  [i/o] Missile list. Launched missile will be inserted to this list.
 * @param type      [in]  Type of a missile.
 * @param screen    [in]  Screen definition.
 * @param launch_x  [in]  X value of launched position.
 * @param launch_y  [in]  Y value of launched position.
 * @param dir_x     [in]  X value of launch direction. Vector (dir_x,dir_y) will be normalized.
 * @param dir_y     [in]  Y value of launch direction. Vector (dir_x,dir_y) will be normalized.
 */
void
missile_launch_three_way( RLgtApp* app, RLlist* missiles, MissileType type, 
  RLrect* screen, double launch_x, double launch_y, double dir_x, double dir_y )
{
	if ( app == NULL || missiles == NULL || screen == NULL ) {
		return;
	}

	{
		Entity missile;
		double angle;
		double vx;
		double vy;
		int angles[] = { 0, 3, 5 };
		int num_of_missiles = 3;
		int i;

		NORMALIZE( dir_x, dir_y );

		if ( fabs( dir_x ) < THRESHOLD && fabs( dir_y ) < THRESHOLD ) {
			return;
		}

		for ( i = 0; i < num_of_missiles; i ++ ) {
			angle = angles[i] * 15.0 / ( 2.0 * PI );
			missile_init( &missile, app, type, screen );
			vx = missile.vmax * ( dir_x * cos( angle ) - dir_y * sin( angle ) );
			vy = missile.vmax * ( dir_x * sin( angle ) + dir_y * cos( angle ) );
			entity_set_position( &missile, launch_x, launch_y );
			entity_set_velocity( &missile, vx, vy );
			entity_set_rotation( &missile, -atan2( vy, vx ) );
			rlib_list_insert( missiles, -1, ( unsigned char*)( &missile ), sizeof( Entity ) );
		}
	}
}


/*!
 * Launches 360 degrees blasting missile.
 *
 * @param app       [in]  GUI toolkit context.
 * @param missiles  [i/o] Missile list. Launched missile will be inserted to this list.
 * @param type      [in]  Type of a missile.
 * @param screen    [in]  Screen definition.
 * @param launch_x  [in]  X value of launched position.
 * @param launch_y  [in]  Y value of launched position.
 */
void
missile_launch_blast( RLgtApp* app, RLlist* missiles, MissileType type, 
      RLrect* screen, double launch_x, double launch_y )
{
	if ( app == NULL || missiles == NULL || screen == NULL ) {
		return;
	}

	{
		Entity missile;
		double angle;
		double vx;
		double vy;
		int num_of_missiles = 24;
		int i;

		for ( i = 0; i < num_of_missiles; i ++ ) {
			angle = i * 15.0 / ( 2.0 * PI );
			missile_init( &missile, app, type, screen );
			vx = missile.vmax * cos( angle );
			vy = missile.vmax * sin( angle );
			entity_set_position( &missile, launch_x, launch_y );
			entity_set_velocity( &missile, vx, vy );
			entity_set_rotation( &missile, -atan2( vy, vx ) );
			rlib_list_insert( missiles, -1, ( unsigned char*)( &missile ), sizeof( Entity ) );
		}
	}
}


/* End of missile.c */
