/*
 * Copyright (c) 2005,2006 Ryo Ichinose
 *
 * This file is part of Space Maze.
 * 
 * Space Maze is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; version 2 of the License.
 * 
 * Space Maze is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with Space Maze; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

/*!
 * @file
 * @brief Handles the space ship entity data and draws it onto the screen.
 *
 * "$Id: spaceship.c,v 1.1.1.1 2006/07/22 18:07:56 ryo Exp $"
 */

#include <math.h>
#include <stdio.h>
#include <string.h>

#include "RLgt.h"
#include "RLsound.h"
#include "entity.h"
#include "explosion.h"
#include "missile.h"
#include "spaceship.h"
#include "resource.h"


#define SPACESHIP_MAX_ENERGY    300
#define SPACESHIP_MAX_MISSILES  10

#define SPACESHIP_ACCT_THRUST   0.1
#define SPACESHIP_ACCT_WHEEL    0.01

#define SPACESHIP_MAX_VELOCITY  2.2
#define SPACESHIP_MIN_VELOCITY  0.005

#define SPACESHIP_MAX_ANGULAR_VELOCITY 0.03
#define SPACESHIP_MIN_ANGULAR_VELOCITY 0.001


static void spaceship_draw_body( Entity* spaceship );
static void spaceship_draw_wing( Entity* spaceship );
static void spaceship_draw_forward_thrust( Entity* spaceship );
static void spaceship_draw_backward_thrust( Entity* spaceship );
static int  spaceship_key_pressed( void* data );
static int  spaceship_key_released( void* data );
static void spaceship_launch_missile( Entity* spaceship );


static inline double 
spaceship_get_acct_thrust()
{
	return SPACESHIP_ACCT_THRUST / entity_get_fps_ratio();
}


static inline double 
spaceship_get_acct_wheel()
{
	return SPACESHIP_ACCT_WHEEL / entity_get_fps_ratio();
}


static inline double 
spaceship_get_max_velocity()
{
	return SPACESHIP_MAX_VELOCITY / entity_get_fps_ratio();
}


static inline double 
spaceship_get_min_velocity()
{
	return SPACESHIP_MIN_VELOCITY / entity_get_fps_ratio();
}


static inline double 
spaceship_get_max_angular_velocity()
{
	return SPACESHIP_MAX_ANGULAR_VELOCITY / entity_get_fps_ratio();
}


static inline double 
spaceship_get_min_angular_velocity()
{
	return SPACESHIP_MIN_ANGULAR_VELOCITY / entity_get_fps_ratio();
}


/*!
 * Initializes space ship entity's data structure.
 *
 * @param spaceship [i/o] The space ship data storage.
 * @param app       [in]  GUI toolkit context.
 * @param screen    [in]  Screen definition on which the space ship should be drawn.
 */
void
spaceship_init( Entity* spaceship, RLgtApp* app, RLsoundContext* sound, RLrect* screen )
{
	if ( spaceship == NULL || app == NULL || sound == NULL || screen == NULL ) {
		return;
	}

	entity_init( spaceship, app, sound, screen );

	spaceship->width             = SPACESHIP_WIDTH;
	spaceship->height            = SPACESHIP_HEIGHT;
	spaceship->energy            = SPACESHIP_MAX_ENERGY;
	spaceship->vmax              = spaceship_get_max_velocity();
	spaceship->decrease_func     = spaceship_decrease_energy;
	spaceship->check_func        = spaceship_check;
	spaceship->update_func       = spaceship_update;
	spaceship->draw_func         = spaceship_draw;
}


/*!
 * Initializes space ship entity's data structure ( demo version for the ending scene ).
 */
void 
spaceship_demo_init( Entity* spaceship, RLgtApp* app, RLsoundContext* sound, RLrect* screen )
{
	spaceship_init( spaceship, app, sound, screen );
	spaceship->vrot = spaceship_get_max_angular_velocity();
}


/*!
 * Sets event handler.
 */
void
spaceship_set_event_handler( Entity* spaceship )
{
	if ( spaceship != NULL && spaceship->gt_app != NULL ) {
		rlib_gt_add_event_handler( spaceship->gt_app, 
			RL_GT_KEY_PRESSED,  spaceship_key_pressed,  ( void* )spaceship );
		rlib_gt_add_event_handler( spaceship->gt_app, 
			RL_GT_KEY_RELEASED, spaceship_key_released, ( void* )spaceship );
	}
}


/*!
 * Decreases space ship enerty with specified amount.
 * If the space ship is blinking because of previous damage,
 * no decreasing is done.
 */
void
spaceship_decrease_energy( Entity* spaceship, int amount )
{
	if ( spaceship == NULL || amount < 0 ) {
		return;
	}

	if ( spaceship->effect_blink_frame_count != 0 ) {
		return;
	}

	rlib_sound_play( spaceship->sound, spaceship->wave_index[ INDEX_WAVE_DAMAGE ], 0 );
	spaceship->energy -= amount;
	spaceship->effect_blink_frame_count = entity_get_framecount_effect_blink();
}


/*!
 * Checks input and computes new position just for collision check.
 * Note that this function does not compute real position.
 * After this, execute spaceship_update() to compute real position.
 * This function returns 0 when no collision check is needed because 
 * spaceship is not alive ...
 *
 * - Checks keyboard event.
 * - If 'd'   key is pressed,  starts rotating clock wise.
 * - If 's'   key is pressed,  starts rotating counter clock wise.
 * - If 'n'   key is pressed,  starts thrusting forward.
 * - If 'm'   key is pressed,  starts thrusting backward.
 * - If SPACE key is pressed,  launch a missile!
 * - If 'd'   key is released, stops rotating clock wise.
 * - If 's'   key is released, stops rotating counter clock wise.
 * - If 'n'   key is released, stops thrusting forward.
 * - If 'm'   key is released, stops thrusting backward.
 *
 * @param spaceship [i/o] The space ship data storage.
 * @param rect      [out] Rectangle that represents new position.
 * @param vx        [out] X value of new velocity.
 * @param vy        [out] Y value of new velocity.
 */
int
spaceship_check( Entity* spaceship, RLrect* rect, double* vx, double* vy )
{
	if ( spaceship == NULL || spaceship->gt_app == NULL || 
	     rect == NULL || vx  == NULL || vy == NULL ) {
		return 0;
	}

	if ( ! entity_is_alive( spaceship ) ) {
		return 0;
	}

	/*
	 * First, determines the value of an accelerlation with its thrusting status.
	 * Then calculates new velocity and adjusts it to be within the maximum velocity.
	 * After that, calculates to-be position at the next frame as a rectangle.
	 */
	if ( spaceship->is_thrusting ) {
		spaceship->ax =   spaceship_get_acct_thrust() * sin( -1.0 * spaceship->prot );
		spaceship->ay = - spaceship_get_acct_thrust() * cos( -1.0 * spaceship->prot );
	}
	else if ( spaceship->is_back_thrusting ) {
		spaceship->ax = - spaceship_get_acct_thrust() * sin( -1.0 * spaceship->prot );
		spaceship->ay =   spaceship_get_acct_thrust() * cos( -1.0 * spaceship->prot );
	}
	else {
		spaceship->ax = spaceship->ay = 0;
	}

	spaceship->vx += spaceship->ax;
	spaceship->vy += spaceship->ay;

	{
		double velocity = sqrt( 
			pow( spaceship->vx, 2 ) + pow( spaceship->vy, 2 ) );

		if ( velocity > spaceship_get_max_velocity() ) {
			spaceship->vx = spaceship_get_max_velocity() * ( spaceship->vx / velocity );
			spaceship->vy = spaceship_get_max_velocity() * ( spaceship->vy / velocity );
		}

		if ( velocity < spaceship_get_min_velocity() ) {
			spaceship->vx = spaceship->vy = 0;
		}
	}

	rect->lefttop_x = spaceship->px + spaceship->vx - spaceship->width  / 2;
	rect->lefttop_y = spaceship->py + spaceship->vy - spaceship->height / 2;
	rect->width     = spaceship->width;
	rect->height    = spaceship->height;

	(*vx) = spaceship->vx;
	(*vy) = spaceship->vy;

	return 1;
}


/*!
 * Commits new position, status, etc. at the next frame.
 */
void
spaceship_update( Entity* spaceship, Entity* dummy )
{
	if ( spaceship == NULL ) {
		return;
	}

	if ( spaceship->is_exploding ) {
		explosion_update( &( spaceship->explosion ), spaceship->px, spaceship->py );

		if ( ! explosion_is_exploding( &( spaceship->explosion ) ) ) {
			spaceship->is_alive     = 0;
			spaceship->is_exploding = 0;
		}
	}
	else if ( spaceship->is_alive ) {
		if ( spaceship->effect_blink_frame_count > 0 ) {
			spaceship->effect_blink_frame_count --;
		}

		spaceship->px   += spaceship->vx;
		spaceship->py   += spaceship->vy;
		spaceship->vrot += spaceship->arot;

		if ( fabs( spaceship->vrot ) > spaceship_get_max_angular_velocity() ) {
			if ( spaceship->vrot > 0 ) {
				spaceship->vrot = spaceship_get_max_angular_velocity();
			}
			else {
				spaceship->vrot = -1.0 * spaceship_get_max_angular_velocity();
			}
		}

		if ( fabs( spaceship->vrot ) < spaceship_get_min_angular_velocity() ) {
			spaceship->vrot = 0;
		}

		spaceship->prot += spaceship->vrot;
	}
}


/*!
 * Draws the space ship entity onto the screen.
 */
void
spaceship_draw( Entity* spaceship )
{
	if ( spaceship == NULL || spaceship->gt_app == NULL ) {
		return;
	}

	if ( spaceship->is_exploding ) {
		explosion_draw( &( spaceship->explosion ) );
	}
	else if ( spaceship->is_alive ) {
		spaceship_draw_body( spaceship );
		spaceship_draw_wing( spaceship );

		if ( spaceship->is_thrusting ) {
			spaceship_draw_forward_thrust( spaceship );
		}
		else if ( spaceship->is_back_thrusting ) {
			spaceship_draw_backward_thrust( spaceship );
		}
	}
}


static void
spaceship_draw_body( Entity* spaceship )
{
#define SPACESHIP_BODY_POINTS_NUM 7

	const double points[ SPACESHIP_BODY_POINTS_NUM ][2] = {
		{   5.0,  20.0 },
		{  10.0, - 5.0 },
		{   5.0, -20.0 },
		{ - 5.0, -20.0 },
		{ -10.0, - 5.0 },
		{ - 5.0,  20.0 },
		{   5.0,  20.0 }
	};

	Point converted_points[ SPACESHIP_BODY_POINTS_NUM ];
	int screen_x;
	int screen_y;
	int i;


	/*
	 * The space ship must be at the center of the playing screen.
	 * Screen position of the other objects like enemies, missiles are 
	 * calculated by SCREEN_POSITION macro though.
	 */
	screen_x = spaceship->screen.lefttop_x + spaceship->screen.width  / 2;
	screen_y = spaceship->screen.lefttop_y + spaceship->screen.height / 2;


	for ( i = 0; i < SPACESHIP_BODY_POINTS_NUM; i ++ ) {
		converted_points[i].x = ( short )( 
			screen_x + (
				points[i][0] * cos( spaceship->prot ) -
				points[i][1] * sin( spaceship->prot ) ) );

		converted_points[i].y = ( short )(
			screen_y - (
				points[i][0] * sin( spaceship->prot ) +
				points[i][1] * cos( spaceship->prot ) ) );
	}


	if ( spaceship->effect_blink_frame_count != 0 &&
	     spaceship->effect_blink_frame_count % 5 == 0 ) {
		rlib_gt_set_foreground_color( spaceship->gt_app, RL_GT_MAX_RGB, 0, 0 );
	}
	else {
		rlib_gt_set_foreground_color( spaceship->gt_app, 
			   RL_GT_MAX_RGB, RL_GT_MAX_RGB, RL_GT_MAX_RGB );
	}

	rlib_gt_fill_polygon( 
		spaceship->gt_app,
		converted_points, 
		SPACESHIP_BODY_POINTS_NUM,
		RL_GT_POLYGON_CONVEX );
}


static void
spaceship_draw_wing( Entity* spaceship )
{
#define SPACESHIP_WING_POINTS_NUM 6

	const double points[ SPACESHIP_WING_POINTS_NUM  ][2] = {
		{ 12.5,   5.0 },
		{ 17.5,   5.0 },
		{ 20.0, -20.0 },
		{ 17.5, -20.0 },
		{ 10.0, - 5.0 },
		{ 12.5,   5.0 },
	};

	Point converted_points[ SPACESHIP_WING_POINTS_NUM ];
	int screen_x;
	int screen_y;
	int i;


	/*
	 * The space ship must be at the center of the playing screen.
	 * Screen position of the other objects like enemies, missiles are 
	 * calculated by SCREEN_POSITION macro though.
	 */
	screen_x = spaceship->screen.lefttop_x + spaceship->screen.width  / 2;
	screen_y = spaceship->screen.lefttop_y + spaceship->screen.height / 2;

	rlib_gt_set_foreground_color( spaceship->gt_app, RL_GT_MAX_RGB, 0, 0 );

	/*
	 * First we draws the right wing of the space ship.
	 * Then, draws the left wing.
	 */
	for ( i = 0; i < SPACESHIP_WING_POINTS_NUM; i ++ ) {
		converted_points[i].x = ( short )( 
			screen_x + (
				points[i][0] * cos( spaceship->prot ) -
				points[i][1] * sin( spaceship->prot ) ) );

		converted_points[i].y = ( short )(
			screen_y - (
				points[i][0] * sin( spaceship->prot ) +
				points[i][1] * cos( spaceship->prot ) ) );
	}

	rlib_gt_fill_polygon( 
		spaceship->gt_app,
		converted_points, 
		SPACESHIP_WING_POINTS_NUM,
		RL_GT_POLYGON_CONVEX );


	for ( i = 0; i < SPACESHIP_WING_POINTS_NUM; i ++ ) {
		converted_points[i].x = ( short )( 
			screen_x + (
				- points[i][0] * cos( spaceship->prot ) -
				  points[i][1] * sin( spaceship->prot ) ) );

		converted_points[i].y = ( short )(
			screen_y - (
				- points[i][0] * sin( spaceship->prot ) +
				  points[i][1] * cos( spaceship->prot ) ) );
	}

	rlib_gt_fill_polygon( 
		spaceship->gt_app,
		converted_points, 
		SPACESHIP_WING_POINTS_NUM, 
		RL_GT_POLYGON_CONVEX );
}


#define SPACESHIP_THRUST_LINE_NUM 3

static void
spaceship_draw_thrust( Entity* spaceship, 
       const double points[ SPACESHIP_THRUST_LINE_NUM ][2][2] );


static void
spaceship_draw_forward_thrust( Entity* spaceship )
{
	const double forward_points[ SPACESHIP_THRUST_LINE_NUM ][2][2] = {
		{ { 0, -20.0 }, { -13.333, -40.0 } },
		{ { 0, -20.0 }, {   0,     -40.0 } },
		{ { 0, -20.0 }, {  13.333, -40.0 } }
	};

	spaceship_draw_thrust( spaceship, forward_points );
}


static void
spaceship_draw_backward_thrust( Entity* spaceship )
{
	const double backward_points[ SPACESHIP_THRUST_LINE_NUM ][2][2] = {
		{ { 0, 20.0 }, { -13.333, 40.0 } },
		{ { 0, 20.0 }, {   0,     40.0 } },
		{ { 0, 20.0 }, {  13.333, 40.0 } }
	};

	spaceship_draw_thrust( spaceship, backward_points );
}


static void
spaceship_draw_thrust( Entity* spaceship, 
       const double points[ SPACESHIP_THRUST_LINE_NUM ][2][2] )
{
	Point converted_points[ SPACESHIP_THRUST_LINE_NUM ][2];
	int screen_x;
	int screen_y;
	int i;
	int j;

	/*
	 * The space ship must be at the center of the playing screen.
	 * Screen position of the other objects like enemies, missiles are 
	 * calculated by SCREEN_POSITION macro though.
	 */
	screen_x = spaceship->screen.lefttop_x + spaceship->screen.width  / 2;
	screen_y = spaceship->screen.lefttop_y + spaceship->screen.height / 2;

	rlib_gt_set_foreground_color( spaceship->gt_app, RL_GT_MAX_RGB, 0, 0 );


	for ( i = 0; i < SPACESHIP_THRUST_LINE_NUM; i ++ ) {
		for ( j = 0; j < 2; j ++ ) {
			converted_points[i][j].x = ( short )( 
				screen_x + (
					points[i][j][0] * cos( spaceship->prot ) -
					points[i][j][1] * sin( spaceship->prot ) ) );

			converted_points[i][j].y = ( short )(
				screen_y - (
					points[i][j][0] * sin( spaceship->prot ) +
					points[i][j][1] * cos( spaceship->prot ) ) );
		}
	}

	for ( i = 0; i < SPACESHIP_THRUST_LINE_NUM; i ++ ) {
		rlib_gt_draw_line( 
			spaceship->gt_app,
			converted_points[i][0].x, 
			converted_points[i][0].y,
			converted_points[i][1].x,
			converted_points[i][1].y );
	}
}


/*!
 * Calculates space ship's new velocity when it hits an enemy.
 */
void
spaceship_reflect( Entity* spaceship, Entity* enemy )
{
	double distance;
	double rx_se;
	double ry_se;
	double velocity_s;
	double velocity_e;
	double px_s;
	double py_s;
	double px_e;
	double py_e;
	double vx_s;
	double vy_s;
	double vx_e;
	double vy_e;

	if ( spaceship == NULL || enemy == NULL ) {
		return;
	}

	entity_get_position( spaceship, &px_s, &py_s );
	entity_get_position( enemy, &px_e, &py_e );

	/*
	 * Calculates a vector to connect center point of the space ship and an enemy,
	 * which direction is from the space ship to the enemy.
	 * Suppose that a reaction occures along this line.
	 */
	rx_se = px_e - px_s;
	ry_se = py_e - py_s;

	distance = sqrt( pow( rx_se, 2 ) + pow( ry_se, 2 ) );

	if ( distance >= THRESHOLD ) { /* To avoid 0 division. */
		rx_se /= distance;
		ry_se /= distance;
	}
	else {
		rx_se = ry_se = 0.0;
	}

	/*
	 * Calculates space ship's new velocity by adding the enemy's velocity
	 * on the same direction of the vector that is already computed above.
	 */
	entity_get_velocity( spaceship, &vx_s, &vy_s );
	entity_get_velocity( enemy, &vx_e, &vy_e );

	velocity_e = sqrt( pow( vx_e, 2 ) + pow( vy_e, 2 ) );

	vx_s += velocity_e * ( -1.0 ) * rx_se;
	vy_s += velocity_e * ( -1.0 ) * ry_se;

	/* 
	 * Then normalizes the space ship's velocity to be within
	 * maximum value, and commits new velocity at the next frame.
	 */
	velocity_s = sqrt( pow( vx_s, 2 ) + pow( vy_s, 2 ) );

	if ( velocity_s >= spaceship->vmax ) {
		vx_s = spaceship->vmax * vx_s / velocity_s;
		vy_s = spaceship->vmax * vy_s / velocity_s;
	}

	entity_set_velocity( spaceship, vx_s, vy_s );
}


static int
spaceship_key_pressed( void* data )
{
	Entity* spaceship = ( Entity* )data;
	RLgtApp* app = spaceship->gt_app;

	if ( rlib_gt_key_pressed( app, RL_GT_KEY_d ) ) {
		spaceship->arot = -1.0 * spaceship_get_acct_wheel();
	}

	if ( rlib_gt_key_pressed( app, RL_GT_KEY_s ) ) {
		spaceship->arot = spaceship_get_acct_wheel();
	}

	if ( rlib_gt_key_pressed( app, RL_GT_KEY_n ) ) {
		spaceship->is_thrusting = 1;
	}

	if ( rlib_gt_key_pressed( app, RL_GT_KEY_m ) ) {
		spaceship->is_back_thrusting = 1;
	}

	if ( rlib_gt_key_pressed( app, RL_GT_KEY_SPACE ) ) {
		rlib_sound_play( spaceship->sound, spaceship->wave_index[ INDEX_WAVE_SHOT ], 0 );
		spaceship_launch_missile( spaceship );
	}

	return 0;
}


static int
spaceship_key_released( void* data )
{
	Entity* spaceship = ( Entity* )data;
	RLgtApp* app = spaceship->gt_app;

	if ( rlib_gt_key_released( app, RL_GT_KEY_d ) ) {
		spaceship->arot = 0;
	}

	if ( rlib_gt_key_released( app, RL_GT_KEY_s ) ) {
		spaceship->arot = 0;
	}

	if ( rlib_gt_key_released( app, RL_GT_KEY_n ) ) {
		spaceship->is_thrusting = 0;
	}

	if ( rlib_gt_key_released( app, RL_GT_KEY_m ) ) {
		spaceship->is_back_thrusting = 0;
	}

	return 0;
}


static void
spaceship_launch_missile( Entity* spaceship )
{
	double dir_x =  0;
	double dir_y = -1;

	if ( spaceship == NULL ) {
		return;
	}

	if ( rlib_list_size( spaceship->missiles ) >= SPACESHIP_MAX_MISSILES ) {
		return;
	}

	dir_x = -1.0 * sin( spaceship->prot );
	dir_y = -1.0 * cos( spaceship->prot );

	missile_launch_single_shot(
		spaceship->gt_app, 
		spaceship->missiles,
		MISSILE_LASER,
		&( spaceship->screen ),
		spaceship->px,
		spaceship->py,
		dir_x,
		dir_y
		);
}


/* End of spaceship.c */
