/*
 * Copyright (c) 2005,2006 Ryo Ichinose
 *
 * This file is part of Space Maze.
 * 
 * Space Maze is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; version 2 of the License.
 * 
 * Space Maze is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with Space Maze; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

/*!
 * @file
 * @brief Common functions for state_xxxx.c.
 *
 * "$Id: state.c,v 1.2 2006/09/21 14:45:19 ryo Exp $"
 */

#ifdef HAVE_CONFIG_H
#  include "config.h"
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#ifdef HAVE_UNISTD_H
#  include <unistd.h>
#  ifdef HAVE_UNLINK
#    define REMOVE_FILE unlink
#  endif
#endif

#include "RLgt.h"
#include "RLlist.h"
#include "RLlogger.h"
#include "RLrect.h"
#include "RLsound.h"
#include "enemy.h"
#include "item.h"
#include "missile.h"
#include "spacemaze.h"
#include "state.h"
#include "resource.h"


#define STATE_FRAME_COUNT_DECREASE_ENERGY 100
#define STATE_FRAME_COUNT_EFFECT_BLINK     50 
#define STATE_FRAME_COUNT_EFFECT_MESSAGE  100


#ifdef HAVE_GETENV
#define GET_FILENAME( filename ) { \
        snprintf( (filename), sizeof( (filename) ), "%s/.spacemaze", getenv( "HOME" ) ); \
}
#endif

static void state_check_collision_with_ship_missiles( 
	SpaceMazeApp* app, Entity* enemy, RLrect* e_rect_ptr );

static void state_remove_dead_entities( SpaceMazeApp* app, RLlist* list );
static Entity* state_get_entity( RLlist* list, int index );


static double f_ratio;


/*!
 * Sets fps ratio ( current fps / ( default fps = 60 ) ).
 */
void
state_set_fps_ratio( double ratio )
{
	f_ratio = ratio;
}


/*!
 * Gets fps ratio ( current fps / ( default fps = 60 ) ).
 */
double
state_get_fps_ratio()
{
	return f_ratio;
}


int
state_get_framecount_decrease_energy()
{
	return ( int )( STATE_FRAME_COUNT_DECREASE_ENERGY * f_ratio );
}


int
state_get_framecount_effect_blink()
{
	return ( int )( STATE_FRAME_COUNT_EFFECT_BLINK * f_ratio );
}


int
state_get_framecount_effect_message()
{
	return ( int )( STATE_FRAME_COUNT_EFFECT_MESSAGE * f_ratio );
}


/*!
 * This function calculates to-be position of all enemies and 
 * spaceship's missiles at the next frame, and checks if there 
 * are any collision between those objects. 
 *
 * If enemies hit each other, calculates some reflection.
 * If a spaceship's missile hits an enemy, the energy of that 
 * enemy will be decreased. And if a spaceship hits an enemy,
 * the energy of the spaceship will be decreased.
 *
 * New position of all enemies will be committed by this function,
 * but not about spaceship's missiles. To commit new position
 * of spaceship's missiles, call state_update_ship_missiles().
 *
 * @param app        [i/o] Application context.
 * @param s_rect_ptr [in]  To-be position of the space ship at the next frame.
 *                         If this pointer is NULL, no collision check between
 *                         space ship and enemies, spaceship and enemy missiles
 *                         will be executed.
 */
void
state_update_enemies( SpaceMazeApp* app, RLrect* s_rect_ptr )
{
	Entity* spaceship = ( s_rect_ptr == NULL ) ? NULL : &( app->spaceship );
	Entity* enemy1;
	Entity* enemy2;
	RLrect e_rect;
	double vx;
	double vy;
	int i;
	int j;

	if ( rlib_list_size( &( app->enemies ) ) == 0 ) {
		state_check_collision_with_ship_missiles( app, NULL, NULL );
		return;
	}

	for ( i = 0; i < rlib_list_size( &( app->enemies ) ); i ++ ) {
		if ( ( enemy1 = state_get_entity( &( app->enemies ), i ) ) == NULL ) {
			continue;
		}

		/*
		 * 1. Checks collision between enemies and calculates reflection.
		 * 2. Checks collision between an enemy and a map block, and calculates reflection.
		 * 3. Checks collision between an enemy and the space ship.
		 * 4. Checks collision between an enemy and a space ship missile.
		 * 5. Commits new position of an enemy at the next frame.
		 */

		for ( j = i + 1 ; j < rlib_list_size( &( app->enemies ) ); j ++ ) {
			if ( ( enemy2 = state_get_entity( &( app->enemies ), j ) ) == NULL ) {
				continue;
			}

			if ( entity_calculate_reflection( enemy1, enemy2 ) ) {
				break;
			}
		}

		if ( entity_check( enemy1, &e_rect, &vx, &vy ) ) {
			unsigned char type;
			if ( ! map_is_movable( &( app->map ), &e_rect, &vx, &vy, &type ) ) {
				entity_set_velocity( enemy1, vx, vy );
			}
		}

		if ( s_rect_ptr != NULL && entity_is_alive( &( app->spaceship ) ) ) {
			if ( entity_is_alive( enemy1 ) && rlib_rect_overlap( s_rect_ptr, &e_rect ) ) {
				if ( ! app->is_debug_on ) {
					entity_calculate_reflection( enemy1, &( app->spaceship ) );
					entity_decrease_energy( &( app->spaceship ), entity_get_strength( enemy1 ) );
					entity_blink( enemy1 );
				}
			}
		}

		state_check_collision_with_ship_missiles( app, enemy1, &e_rect );
		entity_update( enemy1, spaceship );
	}

	state_remove_dead_entities( app, &( app->enemies ) );
}

static void
state_check_collision_with_ship_missiles( SpaceMazeApp* app, Entity* enemy, RLrect* e_rect_ptr )
{
	Entity* missile;
	RLrect m_rect;
	double vx;
	double vy;
	int i;

	for ( i = 0; i < rlib_list_size( &( app->s_missiles ) ); i ++ ) {
		if ( ( missile = state_get_entity( &( app->s_missiles ), i ) ) == NULL ) {
			continue;
		}

		if ( entity_check( missile, &m_rect, &vx, &vy ) ) {
			unsigned char type;
			if ( ! map_is_movable( &( app->map ), &m_rect, &vx, &vy, &type ) ) {
				entity_explode( missile );
			}
		}

		if ( entity_is_alive( enemy ) && rlib_rect_overlap( &m_rect, e_rect_ptr ) ) {
			entity_decrease_energy( enemy, entity_get_strength( missile ) );
			entity_explode( missile );

			/*
			 * entity_key_released() returns 1 only when the enemy "Flyer"
			 * moves into suspend mode for the first time. 
			 * See "src/object/entity.c" and "src/object/enemy/flyer.c" .
			 */
			if ( entity_key_released( enemy ) ) {
				map_show_key_lock( &( app->map ), 0 );
			}
		}
	}

	state_remove_dead_entities( app, &( app->s_missiles ) );
}


/*!
 * This function calculates to-be position of all enemy missiles
 * at the next frame, and checks if there are any collision.
 * If an enemy missile hits the spaceship ( represented by s_rect_ptr ),
 * the energy of the spaceship will be decreased, and that missile
 * disappears.
 *
 * @param app        [i/o] Application context.
 * @param s_rect_ptr [in]  To-be position of the space ship at the next frame.
 *                         If this pointer is NULL, no collision check between
 *                         spaceship and enemy missiles will be executed.
 */
void
state_update_enemy_missiles( SpaceMazeApp* app, RLrect* s_rect_ptr )
{
	Entity* spaceship = ( s_rect_ptr == NULL ) ? NULL : &( app->spaceship );
	Entity* missile;
	RLrect m_rect;
	double vx;
	double vy;
	int i;

	for ( i = 0; i < rlib_list_size( &( app->e_missiles ) ); i ++ ) {
		if ( ( missile = state_get_entity( &( app->e_missiles ), i ) ) == NULL ) {
			continue;
		}

		if ( entity_check( missile, &m_rect, &vx, &vy ) ) {
			unsigned char type;
			if ( ! map_is_movable( &( app->map ), &m_rect, &vx, &vy, &type ) ) {
				entity_explode( missile );
			}
		}

		if ( s_rect_ptr != NULL && entity_is_alive( &( app->spaceship ) ) ) {
			if ( entity_is_alive( missile ) && rlib_rect_overlap( s_rect_ptr, &m_rect ) ) {
				entity_explode( missile );

				if ( ! app->is_debug_on ) {
					entity_decrease_energy( &( app->spaceship ), entity_get_strength( missile ) );
				}
			}
		}

		entity_update( missile, spaceship );
	}

	state_remove_dead_entities( app, &( app->e_missiles ) );
}


/*!
 * Commits to-be position of spaceship's missiles.
 */
void
state_update_ship_missiles( SpaceMazeApp* app )
{
	int i;

	for ( i = 0; i < rlib_list_size( &( app->s_missiles ) ); i ++ ) {
		entity_update( state_get_entity( &( app->s_missiles ), i ), &( app->spaceship ) );
	}
}


static void
state_remove_dead_entities( SpaceMazeApp* app, RLlist* list )
{
	int index = 0;

	while ( 1 ) {
		if ( index >= rlib_list_size( list ) ) {
			break;
		}

		if ( state_get_entity( list, index ) != NULL ) {
			index ++;
			continue;
		}

		rlib_list_remove( list, index );
	}
}


/*!
 * This function checks if the space ship hits items or not.
 * If the spaces ship hits items, make that item disappear
 * and change game status.
 *
 * @param app        [i/o] Application context.
 * @param s_rect_ptr [in]  To-be position of the space ship at the next frame.
 *                         If this pointer is NULL, no collision check between
 *                         space ship and item will be executed.
 */
void
state_update_items( SpaceMazeApp* app, RLrect* s_rect_ptr )
{
	Entity* spaceship = ( s_rect_ptr == NULL ) ? NULL : &( app->spaceship );
	RLrect i_rect;
	double vx;
	double vy;

	if ( entity_check( &( app->key ), &i_rect, &vx, &vy ) ) {
		unsigned char type;
		if ( ! map_is_movable( &( app->map ), &i_rect, &vx, &vy, &type ) ) {
			entity_set_velocity( &( app->key ), vx, vy );
		}
	}

	if ( s_rect_ptr != NULL && entity_is_alive( &( app->spaceship ) ) ) {
		if ( entity_is_alive( &( app->key ) ) && rlib_rect_overlap( s_rect_ptr, &i_rect ) ) {
			int map_level = map_get_current_level( &( app->map ) );
			playinfo_found_key( &( app->playinfo ), map_level );
			map_show_entrance_lock( &( app->map ), 0 );
			entity_explode( &( app->key ) );
			app->play_screen.effect_evacuate_frame_count = state_get_framecount_effect_message();
			app->play_screen.has_key = 1;
			rlib_sound_play( &( app->sound ), app->wave_index[ INDEX_WAVE_GETKEY ], 0 );
		}
	}

	entity_update( &( app->key ), spaceship );
}


/*!
 * Draws item objects.
 * @param app [in] Application context.
 */
void
state_draw_items( SpaceMazeApp* app )
{
	entity_draw( &( app->key ) );
}


/*!
 * Draws enemy objects.
 * @param app [in] Application context.
 */
void
state_draw_enemies( SpaceMazeApp* app )
{
	int i;

	for ( i = 0; i < rlib_list_size( &( app->enemies ) ); i ++ ) {
		entity_draw( state_get_entity( &( app->enemies ), i ) );
	}
}


/*!
 * Draws enemy missile objects.
 * @param app [in] Application context.
 */
void
state_draw_enemy_missiles( SpaceMazeApp* app )
{
	int i;

	for ( i = 0; i < rlib_list_size( &( app->e_missiles ) ); i ++ ) {
		entity_draw( state_get_entity( &( app->e_missiles ), i ) );
	}
}


/*!
 * Draws space ship missile objects.
 * @param [in] Application context.
 */
void
state_draw_spaceship_missiles( SpaceMazeApp* app )
{
	int i;

	for ( i = 0; i < rlib_list_size( &( app->s_missiles ) ); i ++ ) {
		entity_draw( state_get_entity( &( app->s_missiles ), i ) );
	}
}


/*!
 * Save playing record.
 * @param [in] Application context.
 * @retval 0 when successfully loaded, 1 when failed.
 */
int
state_save( SpaceMazeApp* app )
{
	FILE* fp;
	char  filename[255];
	char  logbuf[64];
	int   i;

	GET_FILENAME( filename );

	if ( ( fp = fopen( filename, "w" ) ) == NULL ) {
		snprintf( logbuf, sizeof( logbuf ), "Failed to open save file %s", filename );
		rlib_logger_warn( logbuf );
		return 1;
	}

	for ( i = 0; i <= MAP_MAX_LEVEL; i ++ ) {
		if ( fprintf( fp, "%1d%02d%02d\n", 
			      app->record.keys[i], app->record.mins[i], app->record.secs[i] ) != 6 ) {

			snprintf( logbuf, sizeof( logbuf ), "fwrite() failed." );
			rlib_logger_warn( logbuf );
			fclose( fp );
			REMOVE_FILE( filename );
			return 1;
		}
	}

	fclose( fp );
	return 0;
}


#define IS_VALID_TIME( digit1, digit2 ) \
( '0' <= digit1 && digit1 <= '5' && '0' <= digit2 && digit2 <= '9' ) ? 1 : 0


/*!
 * Loads playing record.
 * @param [in] Application context.
 * @retval 0 when successfully loaded, 1 when failed.
 */
int
state_load( SpaceMazeApp* app )
{
	FILE* fp;
	char  filename[255];
	char  logbuf[128];
	char  buf[8];
	int   i;

	GET_FILENAME( filename );

	if ( ( fp = fopen( filename, "r" ) ) == NULL ) {
		snprintf( logbuf, sizeof( logbuf ), "Failed to open save file %s", filename );
		rlib_logger_warn( logbuf );
		return 1;
	}

	memset( &( app->record ), 0, sizeof( Record ) );

	for ( i = 0; i <= MAP_MAX_LEVEL; i ++ ) {
		if ( fgets( buf, sizeof( buf ), fp ) == NULL ) {
			snprintf( logbuf, sizeof( logbuf ), 
				  "fgets() failed reading %d th level's data. Save data might be broken.", i );
			rlib_logger_warn( logbuf );
			fclose( fp );
			return 1;
		}

		if ( buf[0] == '1' ) {
			app->record.keys[i] = 1;
		}
		else if ( buf[0] != '0' ) {
			snprintf( logbuf, sizeof( logbuf ), 
				  "%d th key data is broken, read value is %c.", i, buf[0] );
			rlib_logger_warn( logbuf );
			fclose( fp );
			return 1;
		}

		if ( IS_VALID_TIME( buf[1], buf[2] ) ) {
			app->record.mins[i] = ( buf[1] - '0' ) * 10 + ( buf[2] - '0' );
		}
		else {
			snprintf( logbuf, sizeof( logbuf ),
				  "%d th time ( minute ) data is broken, read value is %c%c", i, buf[1], buf[2] );
			rlib_logger_warn( logbuf );
			fclose( fp );
			return 1;
		}

		if ( IS_VALID_TIME( buf[3], buf[4] ) ) {
			app->record.secs[i] = ( buf[3] - '0' ) * 10 + ( buf[4] - '0' );
		}
		else {
			snprintf( logbuf, sizeof( logbuf ),
				  "%d th time ( second ) data is broken, read value is %c%c", i, buf[3], buf[4] );
			rlib_logger_warn( logbuf );
			fclose( fp );
			return 1;
		}
	}

#ifdef DEBUG
	printf( "keys = %d %d %d %d\n", app->record.keys[0], app->record.keys[1], 
		app->record.keys[2], app->record.keys[3] );
	printf( "mins = %d %d %d %d\n", app->record.mins[0], app->record.mins[1], 
		app->record.mins[2], app->record.mins[3] );
	printf( "secs = %d %d %d %d\n", app->record.secs[0], app->record.secs[1], 
		app->record.secs[2], app->record.keys[3] );
#endif

	fclose( fp );
	return 0;
}


/*!
 * Draws enemy's information 
 * ( current number of enemies / initial number of enemies ).
 *
 * @param app [in] Application context.
 */
void
state_draw_enemy_info( SpaceMazeApp* app )
{
#define X_LEFTTOP_DIGIT  370
#define Y_LEFTTOP_DIGIT  10
#define DIGIT_WIDTH      26
#define DIGIT_HEIGHT     50

	int current_enemies = rlib_list_size( &( app->enemies ) );
	int digit[2];
	int lt_x = X_LEFTTOP_DIGIT;
	int lt_y = Y_LEFTTOP_DIGIT;
	int i;

	digit[0] = current_enemies / 10;
	digit[1] = current_enemies - digit[0] * 10;

	for ( i = 0; i < 2; i ++ ) {
		rlib_gt_draw_image( 
			&( app->gt ), 
			INDEX_IMAGE_NUMBERS0_0 + digit[i],
			0,
			0, 
			DIGIT_WIDTH,
			DIGIT_HEIGHT,
			lt_x,
			lt_y
			);

		lt_x += DIGIT_WIDTH;
	}

	rlib_gt_set_foreground_color( &( app->gt ), RL_GT_MAX_RGB, RL_GT_MAX_RGB, 0 );
	rlib_gt_draw_line( &( app->gt ), lt_x, lt_y + DIGIT_HEIGHT, lt_x + DIGIT_WIDTH, lt_y ); 
	lt_x += DIGIT_WIDTH;

	digit[0] = app->initial_number_of_enemies / 10;
	digit[1] = app->initial_number_of_enemies - digit[0] * 10;

	for ( i = 0; i < 2; i ++ ) {
		rlib_gt_draw_image( 
			&( app->gt ), 
			INDEX_IMAGE_NUMBERS0_0 + digit[i],
			0,
			0, 
			DIGIT_WIDTH,
			DIGIT_HEIGHT,
			lt_x,
			lt_y
			);

		lt_x += DIGIT_WIDTH;
	}
}


static Entity*
state_get_entity( RLlist* list, int index )
{
	RLlistData* list_item = rlib_list_get( list, index );
	Entity* entity = ( Entity* )( list_item->buf );

	if ( entity_is_alive( entity ) || entity_is_exploding( entity) ) {
		return entity;
	}

	return NULL;
}


/* End of state.c */

