/*
 * Copyright (c) 2005,2006 Ryo Ichinose
 *
 * This file is part of Space Maze.
 * 
 * Space Maze is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; version 2 of the License.
 * 
 * Space Maze is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with Space Maze; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

/*!
 * @file
 * @brief  Handles ending screen.
 *
 * "$Id: state_ending.c,v 1.2 2006/07/24 06:01:17 ryo Exp $"
 */

#include "RLgt.h"
#include "RLsound.h"
#include "bg.h"
#include "spacemaze.h"
#include "state.h"


#define X_CENTER_TOTALTIME      230
#define X_CENTER_DIGIT          370
#define X_CENTER_RANK           250
#define X_CENTER_RANK_LETTER    400

#define Y_CENTER_CONGRATULATE    70
#define Y_CENTER_THANKYOU       120
#define Y_CENTER_TOTALTIME      220
#define Y_CENTER_RANK           320
#define Y_CENTER_GOBACK         430

#define DIGIT_WIDTH  26
#define DIGIT_HEIGHT 50
#define MARGIN       4

#define SPACESHIP_SCREEN_LT_X   430
#define SPACESHIP_SCREEN_LT_Y   270
#define SPACESHIP_SCREEN_WIDTH  100
#define SPACESHIP_SCREEN_HEIGHT 100

#define TIME_FOR_S  ( 11 * 60 )
#define TIME_FOR_A  ( 13 * 60 )
#define TIME_FOR_B  ( 15 * 60 )

static void state_ending_draw_time( SpaceMazeApp* app );
static void state_ending_draw_rank( SpaceMazeApp* app );
static int state_ending_key_pressed( void* data );


/*!
 * Initialize ending screen related variables of application context.
 * @param app [i/o] Application context.
 */
void
state_ending_init( SpaceMazeApp* app )
{
	RLrect screen;

	rlib_sound_stop_all( &( app->sound ) );
	rlib_sound_play( &( app->sound ), app->wave_index[ INDEX_WAVE_CLEAR ], 0 );

	state_save( app );
	state_ending_set_event_handler( app );

	screen.lefttop_x = SPACESHIP_SCREEN_LT_X;
	screen.lefttop_y = SPACESHIP_SCREEN_LT_Y;
	screen.width     = SPACESHIP_SCREEN_WIDTH;
	screen.height    = SPACESHIP_SCREEN_HEIGHT;

	spaceship_demo_init( &( app->spaceship ), &( app->gt ), &( app->sound ), &screen );
}


/*!
 * Sets event handler.
 */
void
state_ending_set_event_handler( SpaceMazeApp* app )
{
	if ( app != NULL ) {
		rlib_gt_clear_event_handler( &( app->gt ) );
		rlib_gt_add_event_handler( &( app->gt ),
			RL_GT_KEY_PRESSED, state_ending_key_pressed, ( void* )app );
	}
}


/*!
 * This function will be called within main loop.
 *
 * - Checks keyboard event ( These will be done within rlib_gt_process_event() ).
 *   - If escape key is pressed, saves game data and changes state to STATE_TITLE.
 * - Updates star object's information and space ship information.
 *
 * @parm app [i/o] Applicatin context.
 * @retval Next game state.
 */
int
state_ending_update( SpaceMazeApp* app )
{
	if ( app == NULL ) {
		return STATE_INVALID;
	}

	if ( rlib_gt_process_event( &( app->gt ) ) != 0 ) {
		return app->next_state;
	}

	app->effect_blink_frame_count ++;
	UPDATE_STARS( app );
	spaceship_update( &( app->spaceship ), NULL );

	return app->game_state;
}


/*!
 * Draws ending screen.
 * @param app [in] Application context.
 */
void
state_ending_draw( SpaceMazeApp* app )
{
	int x_center = SPACEMAZE_SCREEN_WIDTH / 2;
	int y_center;

	if ( app == NULL ) {
		return;
	}

	bg_draw( &( app->gt ) );
	DRAW_STARS( app );

	y_center = Y_CENTER_CONGRATULATE;
	DRAW_IMAGE( app, INDEX_IMAGE_CONGRATULATE, x_center, y_center );

	y_center = Y_CENTER_THANKYOU;
	DRAW_IMAGE( app, INDEX_IMAGE_THANKYOU, x_center, y_center );

	state_ending_draw_time( app );
	state_ending_draw_rank( app );

	y_center = Y_CENTER_GOBACK;
	if ( app->effect_blink_frame_count < state_get_framecount_effect_blink() / 2 ) {
		DRAW_IMAGE( app, INDEX_IMAGE_GO_BACK, x_center, y_center );
	}
	else if ( app->effect_blink_frame_count >= state_get_framecount_effect_blink() ) {
		app->effect_blink_frame_count = 0;
	}

	entity_draw( &( app->spaceship ) );
}


static void
state_ending_draw_time( SpaceMazeApp* app )
{
#define DIGIT_NUM 5
#define OFFSET_DM 10

	int digit[ DIGIT_NUM ];
	int x_center;
	int y_center;
	int mins = 0;
	int secs = 0;
	int index;
	int i;

	for ( i = 0; i < MAP_MAX_LEVEL + 1; i ++ ) {
		mins += app->record.mins[i];
		secs += app->record.secs[i];

		if ( secs >= 60 ) {
			secs -= 60;
			mins ++;
		}

		if ( mins >= 60 ) {
			mins = 59;
		}
	}

	digit[0] = mins / 10;
	digit[1] = mins - digit[0] * 10;
	digit[2] = ':';
	digit[3] = secs / 10;
	digit[4] = secs - digit[3] * 10;

	x_center = X_CENTER_TOTALTIME;
	y_center = Y_CENTER_TOTALTIME;

	DRAW_IMAGE( app, INDEX_IMAGE_TOTALTIME, x_center, y_center );
	x_center = X_CENTER_DIGIT;

	for ( i = 0; i < DIGIT_NUM; i ++ ) {
		if ( digit[i] == ':' ) {
			index = OFFSET_DM;
		}
		else {
			index = digit[i];
		}

		rlib_gt_draw_image(
			&( app->gt ),
			INDEX_IMAGE_NUMBERS3_0 + index,
			0,
			0,
			DIGIT_WIDTH,
			DIGIT_HEIGHT,
			x_center + MARGIN,
			y_center - DIGIT_HEIGHT / 2
			);

		x_center += DIGIT_WIDTH;
	}

}


static void
state_ending_draw_rank( SpaceMazeApp* app )
{
	int x_center;
	int y_center;
	int image_index;
	int totals = 0;
	int i;

	x_center = X_CENTER_RANK;
	y_center = Y_CENTER_RANK;

	DRAW_IMAGE( app, INDEX_IMAGE_RANK, x_center, y_center );

	for ( i = 0; i < MAP_MAX_LEVEL + 1; i ++ ) {
		totals += app->record.mins[i] * 60;
		totals += app->record.secs[i];
	}

	if ( totals <= TIME_FOR_S ) {
		image_index = INDEX_IMAGE_RANK_S;
	}
	else if ( totals <= TIME_FOR_A ) {
		image_index = INDEX_IMAGE_RANK_A;
	}
	else if ( totals <= TIME_FOR_B ) {
		image_index = INDEX_IMAGE_RANK_B;		
	}
	else {
		image_index = INDEX_IMAGE_RANK_C;
	}

	x_center = X_CENTER_RANK_LETTER;
	DRAW_IMAGE( app, image_index, x_center, y_center );
}


static int
state_ending_key_pressed( void* data )
{
	SpaceMazeApp* app = ( SpaceMazeApp* )data;

	if ( rlib_gt_key_pressed( &( app->gt ), RL_GT_KEY_ESC ) ) {
		rlib_sound_play( &( app->sound ), app->wave_index[ INDEX_WAVE_ESC ], 0 );
		state_title_init( app );
		app->next_state = STATE_TITLE;
		return 1;
	}

	return 0;
}


/* End of state_ending.c */
