/*
 * Copyright (c) 2003-2006 Ryo Ichinose
 *
 * This file is part of Space Maze.
 * 
 * Space Maze is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; version 2 of the License.
 * 
 * Space Maze is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with Space Maze; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

/*!
 * @file
 * @brief List functions.
 *
 * $Id: RLlist.c,v 1.1.1.1 2006/07/22 18:07:56 ryo Exp $
 */

#include <stdio.h>
#include <string.h>

#include "RLerrno.h"
#include "RLlogger.h"
#include "RLlist.h"


static void rlib_list_inner_insert( RLlistData* start, RLlistData* new, int index );

static void rlib_list_inner_insert_by_func( RLlistData* start, RLlistData* new, 
	int ( *func )( unsigned char* new, unsigned char* old ) );

static RLlistData* rlib_list_inner_remove( RLlistData* start, int index );


/*!
 * Initializes a list. "list_ptr" must not be NULL.
 * When "list_ptr" is NULL, this function does nothing.
 *
 * @param list_ptr [i/o] List data struct. Must not be NULL.
 */
void 
rlib_list_init( RLlist* list_ptr )
{
	int i;

	if ( list_ptr == NULL ) {
		return;
	}

	list_ptr->data_num       = 0;
	list_ptr->start.next_ptr = NULL;
	list_ptr->empty.next_ptr = &( list_ptr->data[0] );

	memset( list_ptr->data, 0, sizeof( list_ptr->data ) );

	for ( i = 0; i < RLLIST_MAX_DATA_NUM - 1; i ++ ) {
		list_ptr->data[i].index    = i;
		list_ptr->data[i].next_ptr = &( list_ptr->data[ i + 1 ] );
	}

	list_ptr->data[ RLLIST_MAX_DATA_NUM - 1 ].index    = RLLIST_MAX_DATA_NUM - 1;
	list_ptr->data[ RLLIST_MAX_DATA_NUM - 1 ].next_ptr = NULL;

	return;
}


/*
 * Checks if the specified list is full or not.
 *
 * @param  list_ptr [in] List data struct. Must not be NULL.
 *
 * @retval 1  Specified list is full.
 * @retval 0  Specified list is not full.
 * @retval RLIB_EFAULT "list_ptr" is NULL.
 */
int 
rlib_list_is_full( const RLlist* list_ptr )
{
	if ( list_ptr == NULL ) {
		return RLIB_EFAULT;
	}

	if ( list_ptr->data_num == RLLIST_MAX_DATA_NUM ) {
		return 1;
	}

	return 0;
}


/*!
 * Checks if the specified list is empty or not.
 *
 * @param list_ptr [in] List data struct. Must not be NULL.
 *
 * @retval  1  Specified list is empty.
 * @retval  0  Specified list is not empty.
 * @retval  RLIB_EFAULT "list_ptr" is NULL.
 */
int 
rlib_list_is_empty( const RLlist* list_ptr )
{
	if ( list_ptr == NULL ) {
		return RLIB_EFAULT;
	}

	if ( list_ptr->data_num == 0 ) {
		return 1;
	}

	return 0;
}


/*!
 * Checks if the data specified by "target" is contained in "list_ptr".
 * Start searching from the head of the list. "func" should return 
 * 0 when "obj1" and "obj2" is considered to be equal.
 *
 * @param list_ptr [in] List data struct. Must not be NULL.
 * @param target   [in] Target data to search from list. Must not be NULL.
 * @param func     [in] Search function. Must not be NULL.
 *
 * @retval  RLIB_ENOENT  "target" is not contained in "list_ptr".
 * @retval  RLIB_EFAULT  "list_ptr" or "target" or "func" is NULL.
 * @retval  else the index of the corresponding data is returned.
 */
int
rlib_list_contains( RLlist* list_ptr, unsigned char* target, 
	int ( *func )( unsigned char* obj1, unsigned char* obj2 ) )
{
	RLlistData* prev;
	RLlistData* curr;
	int cnt = 0;

	if ( list_ptr == NULL || target == NULL || func == NULL ) {
		return RLIB_EFAULT;
	}

	prev = &( list_ptr->start );
	curr = prev->next_ptr;

	while ( curr != NULL ) {
		if ( func( target, curr->buf ) == 0 ) {
			return cnt;
		}

		prev = curr;
		curr = curr->next_ptr;
		cnt ++;
	}

	return RLIB_ENOENT;
}


/*!
 * This function returns the number of data in the list.
 *
 * @param list_ptr [in] List data struct. Must not be NULL.
 *
 * @retval RLIB_EFAULT "list_ptr" is NULL.
 * @retval else the number of data in the specified list.
 */
int
rlib_list_size( const RLlist* list_ptr )
{
	if ( list_ptr == NULL ) {
		return RLIB_EFAULT;
	}

	if ( rlib_list_is_empty( list_ptr ) ) {
		return 0;
	}

	return list_ptr->data_num;
}


/*!
 * This function returns a pointer of data in list specified 
 * by "index", which remains in list until rlib_list_remove() 
 * is called. When no data available, this function returns NULL.
 * Also returns NULL when list_ptr is NULL.
 *
 * @param list_ptr [in] List data struct. Must not be NULL.
 * @param index    [in] Index of data to get from the list.
 *
 * @return A pointer represents the specified index. 
 *         NULL when no data is available.
 */
RLlistData*
rlib_list_get( RLlist* list_ptr, int index )
{
	RLlistData* previous;
	RLlistData* current;
	int count = 0;
	int size;

	if ( list_ptr == NULL ) {
		return NULL;
	}

	size = rlib_list_size( list_ptr );

	if ( size == 0 || index >= size ) {
		return NULL;
	}

	previous = &( list_ptr->start );
	current  = previous->next_ptr;

	while ( current != NULL ) {
		if ( count == index ) {
			return current;
		}

		previous = current;
		current = current->next_ptr;
		count ++;
	}

	return NULL;
}


/*!
 * Adds data specified by "buf" ( and length is "len" in bytes )
 * to list at the position specified by "index".
 * If "index" is less than 0 or larger than the last index 
 * of a data in the list, new data will be added to the last 
 * of the list.
 *
 * @param list_ptr [i/o] List data struct. Must not be NULL.
 * @param index    [in]  Index to insert new data.
 * @param buf      [in]  Data buffer to insert into the list. Must not be NULL.
 * @param len      [in]  Length of data in "buf".
 *
 * @retval  0           Successfully processed.
 * @retval  RLIB_ENOSPC No more data can be added.
 * @retval  RLIB_EFAULT "list_ptr" or "buf" is NULL.
 */
int 
rlib_list_insert( RLlist* list_ptr, int index, unsigned char* buf, int len )
{
	RLlistData* target;
	int length_to_copy;

	if ( list_ptr == NULL || buf == NULL ) {
		return RLIB_EFAULT;
	}

	if ( rlib_list_is_full( list_ptr ) ) {
		return RLIB_ENOSPC;
	}

	if ( len > RLLIST_MAX_DATA_LEN ) {
		length_to_copy = RLLIST_MAX_DATA_LEN;
	}
	else {
		length_to_copy = len;
	}

	target = rlib_list_inner_remove( &( list_ptr->empty ), 0 );
	target->len = length_to_copy;
	memcpy( &( target->buf[0] ), buf, length_to_copy );
	rlib_list_inner_insert( &( list_ptr->start ), target, index );
	list_ptr->data_num ++;

	return 0;
}


/*!
 * Adds data specified by "buf" ( and length is "len" in bytes )
 * to the list. Position of the data will be determined by "func".
 * Starts searching from the head of the list. "func" should return 
 * 1 when "new" should be inserted before "old" exists in the list.
 * If "func" is NULL, new data will be added to the last of the list.
 *
 * @param list_ptr [i/o] List data struct. Must not be NULL.
 * @param buf      [in]  Data buffer to insert into the list. Must not be NULL.
 * @param len      [in]  Length of data in "buf".
 * @param func     [in]  Function to determine the position to insert.
 *
 * @retval  0           Successfully processed.
 * @retval  RLIB_ENOSPC No more data can be added.
 * @retval  RLIB_EFAULT "list_ptr" or "buf" is NULL.
 */
int
rlib_list_insert_by_func( RLlist* list_ptr, unsigned char* buf, int len,
	int ( *func )( unsigned char* new, unsigned char* old ) )
{
	RLlistData* target;
	int length_to_copy;

	if ( list_ptr == NULL || buf == NULL ) {
		return RLIB_EFAULT;
	}

	if ( rlib_list_is_full( list_ptr ) ) {
		return RLIB_ENOSPC;
	}

	if ( len > RLLIST_MAX_DATA_LEN ) {
		length_to_copy = RLLIST_MAX_DATA_LEN;
	}
	else {
		length_to_copy = len;
	}

	target = rlib_list_inner_remove( &( list_ptr->empty ), 0 );
	target->len = length_to_copy;
	memcpy( &( target->buf[0] ), buf, length_to_copy );
	rlib_list_inner_insert_by_func( &( list_ptr->start ), target, func );
	list_ptr->data_num ++;

	return 0;
}


/*!
 * Removes the specified data from the list.
 *
 * @param list_ptr [i/o] List data struct. Must not be NULL.
 * @param index    [in]  Index of the data to be removed.
 *
 * @retval  0           Successfully processed.
 * @retval  RLIB_ENOENT Specified data is not available.
 * @retval  RLIB_EFAULT "list_ptr" is NULL.
 */
int
rlib_list_remove( RLlist* list_ptr, int index )
{
	RLlistData* target;
	int size;

	if ( list_ptr == NULL ) {
		return RLIB_EFAULT;
	}

	size = rlib_list_size( list_ptr );

	if ( size == 0 || index >= size ) {
		return RLIB_ENOENT;
	}

	target = rlib_list_inner_remove( &( list_ptr->start ), index );
	rlib_list_inner_insert_by_func( &( list_ptr->empty ), target, NULL );
	list_ptr->data_num --;

	return 0;
}


/*!
 * Removes all of the data from the list.
 * @param list_ptr [i/o] List data struct. Must not be NULL.
 */
void
rlib_list_clear( RLlist* list_ptr )
{
	if ( list_ptr == NULL ) {
		return;
	}

	rlib_list_init( list_ptr );
	return;
}


/*!
 * Prints the contents of the specified list.
 * @param list_ptr [in] List data struct. Must not be NULL.
 */
void 
rlib_list_print( const RLlist* list_ptr )
{
	RLlistData* current;
	char* funcname = "rlib_list_print";
	char  logmsg[256];
	int i;

	if ( list_ptr == NULL ) {
		snprintf( logmsg, sizeof( logmsg ), 
		  "%-13s : NULL pointer specified to %s", __FILE__, funcname );
		rlib_logger_warn( logmsg );
		return;
	}

	snprintf( logmsg, sizeof( logmsg ), 
	  "%-13s : Print contents of specified list.", __FILE__ );
	rlib_logger_debug( logmsg );

	snprintf( logmsg, sizeof( logmsg ), 
	  "%-13s :   Data num = %d", __FILE__, list_ptr->data_num );
	rlib_logger_debug( logmsg );

	for ( i = 0; i < RLLIST_MAX_DATA_NUM; i ++ ) {
		snprintf( logmsg, sizeof( logmsg ), 
		  "%-13s :   Index = %d", __FILE__, list_ptr->data[i].index );

		rlib_logger_debug( logmsg );
		snprintf( logmsg, sizeof( logmsg ), "%-13s :   Dump %d bytes ...", 
		  __FILE__, list_ptr->data[ i ].len );

		rlib_logger_debug( logmsg );

		rlib_logger_hexdump( RLLOGGER_LEVEL_DEBUG,
		  ( unsigned char* )( &( list_ptr->data[ i ].buf[0] ) ),
		  list_ptr->data[ i ].len );
	}

	snprintf( logmsg, sizeof( logmsg ), 
	  "%-13s :   Print index in list.", __FILE__ );

	rlib_logger_debug( "" );
	rlib_logger_debug( logmsg );

	current = list_ptr->start.next_ptr;

	while ( current != NULL ) {
		snprintf( logmsg, sizeof( logmsg ), 
		  "%-13s :   index = %d", __FILE__, current->index );
		rlib_logger_debug( logmsg );
		current = current->next_ptr;
	}

	snprintf( logmsg, sizeof( logmsg ), 
	  "%-13s :   Print index in empty list.", __FILE__ );
	rlib_logger_debug( "" );
	rlib_logger_debug( logmsg );

	current = list_ptr->empty.next_ptr;

	while ( current != NULL ) {
		snprintf( logmsg, sizeof( logmsg ), 
		  "%-13s :   index = %d", __FILE__, current->index );
		rlib_logger_debug( logmsg );
		current = current->next_ptr;
	}

	snprintf( logmsg, sizeof( logmsg ), "%-13s : End.", __FILE__ );
	rlib_logger_debug( "" );
	rlib_logger_debug( logmsg );

	return;
}


static void 
rlib_list_inner_insert( RLlistData* start, RLlistData* new, int index )
{
	RLlistData* previous = start;
	RLlistData* current  = start->next_ptr;
	int count = 0;

	while ( current != NULL ) {
		if ( count == index ) {
			previous->next_ptr = new;
			new->next_ptr = current;
			return;
		}

		previous = current;
		current  = current->next_ptr;
		count ++;
	}

	previous->next_ptr = new;
	new->next_ptr = NULL;
	return;
}


static void 
rlib_list_inner_insert_by_func( RLlistData* start, RLlistData* new,
	int ( *func )( unsigned char* new, unsigned char* old ) )
{
	RLlistData* previous = start;
	RLlistData* current  = start->next_ptr;

	while ( current != NULL ) {
		if ( func != NULL && func( new->buf, current->buf ) == 1 ) {
			previous->next_ptr = new;
			new->next_ptr = current;
			return;
		}

		previous = current;
		current  = current->next_ptr;
	}

	previous->next_ptr = new;
	new->next_ptr = NULL;
	return;
}


static RLlistData*
rlib_list_inner_remove( RLlistData* start, int index )
{
	RLlistData* previous = start;
	RLlistData* current  = start->next_ptr;
	int count = 0;

	while ( current != NULL ) {
		if ( count == index ) {
			previous->next_ptr = current->next_ptr;
			return current;
		}

		previous = current;
		current  = current->next_ptr;
		count ++;
	}

	return NULL;
}


/* End of RLlist.c */
