/*
 * Copyright (c) 2005,2006 Ryo Ichinose
 *
 * This file is part of Space Maze.
 * 
 * Space Maze is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; version 2 of the License.
 * 
 * Space Maze is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with Space Maze; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

/*!
 * @file
 * @brief Logging functions.
 *
 * $Id: RLlogger.c,v 1.1.1.1 2006/07/22 18:07:56 ryo Exp $
 */

#include <stdio.h>
#include "RLlogger.h"


/*!
 * Logs fatal message.
 * @param message [i] Log message.
 */
void
rlib_logger_fatal( const char* message )
{
	if ( message == NULL ) {
		return;
	}

	fprintf( stderr, "[FATAL]  : %s\n", message );
}


/*!
 * Logs alert message.
 * @param message [i] Log message.
 */
void
rlib_logger_alert( const char* message )
{
	if ( message == NULL ) {
		return;
	}

	fprintf( stderr, "[ALERT]  : %s\n", message );
}


/*!
 * Logs crit message.
 * @param message [i] Log message.
 */
void
rlib_logger_crit( const char* message )
{
	if ( message == NULL ) {
		return;
	}

	fprintf( stderr, "[CRIT]   : %s\n", message );
}


/*!
 * Logs error message.
 * @param message [i] Log message.
 */
void
rlib_logger_error( const char* message )
{
	if ( message == NULL ) {
		return;
	}

	fprintf( stderr, "[ERROR]  : %s\n", message );
}


/*!
 * Logs warn message.
 * @param message [i] Log message.
 */
void
rlib_logger_warn( const char* message )
{
	if ( message == NULL ) {
		return;
	}

	fprintf( stderr, "[WARN]   : %s\n", message );
}


/*!
 * Logs notice message.
 * @param message [i] Log message.
 */
void
rlib_logger_notice( const char* message )
{
	if ( message == NULL ) {
		return;
	}

	fprintf( stderr, "[NOTICE] : %s\n", message );
}


/*!
 * Logs debug message.
 * @param message [i] Log message.
 */
void
rlib_logger_info( const char* message )
{
	if ( message == NULL ) {
		return;
	}

	fprintf( stderr, "[INFO]   : %s\n", message );
}


/*!
 * Logs debug message.
 * @param message [i] Log message.
 */
void
rlib_logger_debug( const char* message )
{
	if ( message == NULL ) {
		return;
	}

	fprintf( stderr, "[DEBUG]  : %s\n", message );
}


/*!
 * Logs hex dump of specified data.
 *
 * @param level [i] Log level.
 * @param data  [i] Data buffer to dump.
 * @param len   [i] Length of the data to dump.
 */
void 
rlib_logger_hexdump( char level, const unsigned char* data, int len )
{
	void (*func)( const char* ) = NULL;

	if ( data == NULL || len <= 0 ) {
		return;
	}

	switch ( level ) {
		case RLLOGGER_LEVEL_FATAL:
			func = rlib_logger_fatal;
			break;

		case RLLOGGER_LEVEL_ALERT:
			func = rlib_logger_alert;
			break;

		case RLLOGGER_LEVEL_CRIT:
			func = rlib_logger_crit;
			break;

		case RLLOGGER_LEVEL_ERROR:
			func = rlib_logger_error;
			break;

		case RLLOGGER_LEVEL_WARN:
			func = rlib_logger_warn;
			break;

		case RLLOGGER_LEVEL_NOTICE:
			func = rlib_logger_notice;
			break;

		case RLLOGGER_LEVEL_INFO:
			func = rlib_logger_info;
			break;

		case RLLOGGER_LEVEL_DEBUG:
			func = rlib_logger_debug;
			break;

		default:
			func = rlib_logger_debug;
			break;
	}

	{
		char logmsg[256];
		int  pos;
		int  i;
		int  j;

		snprintf( logmsg, sizeof( logmsg ), "%-13s : Dumping %d bytes.", __FILE__, len );
		func( "" );
		func( logmsg );

		for ( i = 0; i < len; i += 16 ) {
			pos = snprintf( logmsg, sizeof( logmsg ), "%-13s : %08x : ", __FILE__, i );

			for ( j = i; j < i + 16 && j < len; j ++ ) {
				pos += snprintf( logmsg + pos, sizeof( logmsg ) - pos, "%02x", data[j] );

				if ( j % 2 == 1 ) {
					pos += snprintf( logmsg + pos, sizeof( logmsg ) - pos, " " );
				}
			}

			if ( j == len && len % 16 != 0 ) {
				for ( j = 0; j < 40 - ( len % 16 ) * 2.5; j ++ ) {
					pos += snprintf( logmsg + pos, sizeof( logmsg ) - pos, " " );
				}
			}

			pos += snprintf( logmsg + pos, sizeof( logmsg ) - pos, ": " );

			for ( j = i; j < i + 16 && j < len; j ++ ) {
				if ( ( data[j] >= 0x20 ) && ( data[j] <= 0x7e ) ) {
					pos += snprintf( logmsg + pos, sizeof( logmsg ) - pos, "%c", data[j] );
				}
				else { 
					pos += snprintf( logmsg + pos, sizeof( logmsg ) - pos, "." );
				}
			}

			func( logmsg );
		}

		func( "" );
	}
}


/* End of RLlogger.c */
