/*
 * Copyright (c) 2002-2007 Ryo Ichinose
 *
 * This file is part of Space Maze.
 * 
 * Space Maze is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; version 2 of the License.
 * 
 * Space Maze is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with Space Maze; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

/*!
 * @file
 * @brief Utility functions for DSP device using OSS API.
 *
 * $Id: RLdsp.c,v 1.2 2007/11/04 11:41:12 ryo Exp $
 */

#include <errno.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include <linux/soundcard.h>

#include "RLerrno.h"
#include "RLdsp.h"


/*!
 * Gets the current parameter of the specified "fd". 
 * "fd" should be the file descriptor of /dev/dsp.
 * "prms" will be used to save the queried parameters,
 * so "prms" must not be NULL pointer.
 *
 * @param fd   [in]  File descriptor of /dev/dsp.
 * @param prms [i/o] DSP parameter struct. Must not be NULL.
 * @param err  [out] errno will be saved. Must not be NULL.
 *
 * @retval  0             Successfully processed. 
 *                        "prms" will be set to queried parameters.
 *
 * @retval  RLIB_ESYSCALL System call failed.
 *                        "err" will be set to errno, so "err" must not be NULL.
 *
 * @retval  RLIB_EFAULT   "prms" or "err" is NULL.
 */
int 
rlib_dsp_query_current( int fd, RLdspPrms* prms, int* err )
{
	int rtn;

	if ( prms == NULL || err == NULL ) {
		return RLIB_EFAULT;
	}


	/* Gets audio format */
	prms->format = AFMT_QUERY;
	if ( ( rtn = ioctl( fd, SOUND_PCM_SETFMT, &( prms->format ) ) ) == -1 ) {
		*err = errno;
		return RLIB_ESYSCALL;
	}

	/* 
	 * Gets sampling bit size
	 */
	if ( ( rtn = ioctl( fd, SOUND_PCM_READ_BITS, &( prms->bits ) ) ) == -1 ) {
		*err = errno;
		return RLIB_ESYSCALL;
	}

	/* 
	 * Gets sampling rate 
	 */
	if ( ( rtn = ioctl( fd, SOUND_PCM_READ_RATE, &( prms->rate ) ) ) == -1 ) {
		*err = errno;
		return RLIB_ESYSCALL;
	}

	/* 
	 * Gets number of channel
	 */
	if ( ( rtn = ioctl( fd, SOUND_PCM_READ_CHANNELS, &( prms->channel ) ) ) == -1 ) {
		*err = errno;
		return RLIB_ESYSCALL;
	}

	/* 
	 * Gets buffer size in device driver
	 */
	if ( ( rtn = ioctl( fd, SNDCTL_DSP_GETBLKSIZE, &( prms->blksize ) ) ) == -1 ) {
		*err = errno;
		return RLIB_ESYSCALL;
	}

	return 0;
}


/*!
 * Sets the basic parameters of "fd" to the given parameters.
 * "fd" sould be the file descriptor of "/dev/dsp". 
 *
 * Value of "prms" will be used for ioctl(), so "prms" must not 
 * be NULL and needs to be set to the desired value.
 *
 * @param fd   [in]  File descriptor of /dev/dsp.
 * @param prms [i/o] DSP parameter struct. Must not be NULL.
 * @param err  [out] errno will be saved. Must not be NULL.
 *
 * @retval  0             Successfully processed. 
 *
 * @retval  RLIB_ESYSCALL System call failed.
 *                        "err" will be set to errno, so "err" must not be NULL.
 *
 * @retval  RLIB_EFAULT   "prms" or "err" is NULL.
 */
int 
rlib_dsp_control( int fd, RLdspPrms* prms, int* err )
{
	int i, bit, val, rtn;

	if ( prms == NULL || err == NULL ) {
		return RLIB_EFAULT;
	}

	/*
	 * Fragment size must be set before all the other
	 * ioctl commands ( OSS programmer's guide ).
	 */

	/* 
	 * if fragment is set to zero, 
	 * it means we should not change fragment settings 
	 */
	if ( prms->fragment != 0 ) {
		/* 
		 * Before execute ioctl( fd, SNDCTL_DSP_SETFRAGMENT, ... ),
		 *       we must calculate argment value from specified fragment. 
		 */
		for ( i = 14, bit = 0x4000; bit > 0; bit >>= 1, i -- ) {
			if ( prms->fragment >= bit ) {
				break;
			}
		}

		val = ( ( prms->numfrags & 0xffff ) << 16 ) | i;

		/* 
		 * Sets size of device driver's fragment 
		 */
		if ( ( rtn = ioctl( fd, SNDCTL_DSP_SETFRAGMENT, &val ) ) == -1 ) {
			*err = errno;
			return RLIB_ESYSCALL;
		}
	}

	/* Sets audio format */
	if ( ( rtn = ioctl( fd, SNDCTL_DSP_SETFMT, &( prms->format ) ) ) == -1 ) {
		*err = errno;
		return RLIB_ESYSCALL;
	}

	/* 
	 * Sets samling bit size 
	 * <== this is determined by format, so this commented out block
	 * is useless. 
	 */
#if 0
	if ( ( rtn = ioctl( fd, SOUND_PCM_WRITE_BITS, &( prms->bits ) ) ) == -1 ) {
		*err = errno;
		return RLIB_ESYSCALL;
	}
#endif

	/* Sets sampling rate */
	if ( ( rtn = ioctl( fd, SOUND_PCM_WRITE_RATE, &( prms->rate ) ) ) == -1 ) {
		*err = errno;
		return RLIB_ESYSCALL;
	}

	/* Sets number of channel */
	if ( ( rtn = ioctl( fd, SOUND_PCM_WRITE_CHANNELS, &( prms->channel ) ) ) == -1 ) {
		*err = errno;
		return RLIB_ESYSCALL;
	}

	return 0;
}


/*! 
 * Checks supported audio format of specified "fd". 
 * "fd" should be the file descriptor of "/dev/dsp".
 * Supported format will be saved to "fmts", so "fmts" must not be NULL pointer.
 *
 * @param fd   [in]  File descriptor of /dev/dsp.
 * @param fmts [out] Supported format will be saved. Must not be NULL.
 * @param err  [out] This will be set to errno. Must not be NULL.
 *
 * @retval  0             Successfully processed. 
 *
 * @retval  RLIB_ESYSCALL System call failed.
 *                        "err" will be set to errno, so "err" must not be NULL.
 *
 * @retval  RLIB_EFAULT   "fmts" or "err" is NULL.
 */
int 
rlib_dsp_query_format( int fd, RLdspFormat* fmts, int* err )
{
	int fmt_tbl[ RLDSP_NUM_FORMAT ] = {
		AFMT_MU_LAW,
		AFMT_A_LAW,
		AFMT_IMA_ADPCM,
		AFMT_U8,
		AFMT_S16_LE,
		AFMT_S16_BE,
		AFMT_S8,
		AFMT_U16_LE,
		AFMT_U16_BE,
		AFMT_MPEG
	};

	int formats, rtn, i;

	if ( fmts == NULL || err == NULL ) {
		return RLIB_EFAULT;
	}

	/* Gets supported formats */
	if ( ( rtn = ioctl( fd, SOUND_PCM_GETFMTS, &formats ) ) == -1 ) {
		*err = errno;
		return RLIB_ESYSCALL;
	}

	for ( i = 0; i < RLDSP_NUM_FORMAT; i ++ ) {
		( formats & fmt_tbl[i] ) ? 
			( fmts->fmts[i] = 1 ) : ( fmts->fmts[i] = 0 );
	}

	return 0;
}


/*! 
 * Checks capabilities of specified "fd". 
 * "fd" should be the file descriptor of "/dev/dsp".
 * Capabilities will be saved to "caps", so "caps" must not be NULL pointer.
 *
 * @param fd   [in]  File descriptor of /dev/dsp.
 * @param caps [out] Capabilities will be saved. Must not be NULL.
 * @param err  [out] This will be set to errno. Must not be NULL.
 *
 * @retval  0             Successfully processed. 
 *
 * @retval  RLIB_ESYSCALL System call failed.
 *                        "err" will be set to errno, so "err" must not be NULL.
 *
 * @retval  RLIB_EFAULT   "caps" or "err" is NULL.
 */
int 
rlib_dsp_query_caps( int fd, RLdspCaps* caps, int* err )
{
	int caps_tbl[ RLDSP_NUM_CAPS ] = {
		DSP_CAP_DUPLEX,
		DSP_CAP_REALTIME,
		DSP_CAP_BATCH,
		DSP_CAP_COPROC,
		DSP_CAP_TRIGGER,
		DSP_CAP_MMAP,
#if 0
/*
 *  Following two constants seems not to be supported
 *  under my system
 */
		DSP_CAP_MULTI,
		DSP_CAP_BIND
#endif
	};

	int rtn, capabilities, i;

	if ( caps == NULL || err == NULL ) {
		return RLIB_EFAULT;
	}


	/* Gets dsp capabilities */
	if ( ( rtn = ioctl( fd, SNDCTL_DSP_GETCAPS, &capabilities ) ) == -1 ) {
		*err = errno;
		return RLIB_ESYSCALL;
	}

	memset( caps, 0, sizeof( RLdspCaps ) );
	caps->revision = capabilities & DSP_CAP_REVISION;

	for ( i = 0; i < RLDSP_NUM_CAPS; i ++ ) {
		if ( capabilities & caps_tbl[i] ) {
			caps->caps[i] = 1;
		}
		else {
			caps->caps[i] = 0;
		}
	}

	return 0;
}


/* end of RLdsp.c */
