/*
 * Copyright (c) 2002-2007 Ryo Ichinose
 *
 * This file is part of Space Maze.
 * 
 * Space Maze is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; version 2 of the License.
 * 
 * Space Maze is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with Space Maze; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

/*!
 * @file
 * @brief Header file of utility functions for DSP device using OSS API.
 *
 * $Id: RLdsp.h,v 1.2 2007/11/04 11:41:12 ryo Exp $
 */

#ifndef ___RLDSP_H_INCLUDED_
#define ___RLDSP_H_INCLUDED_


/*!
 * Number of audio format.
 */
#define RLDSP_NUM_FORMAT 10


/*!
 * Number of audio device capabilities.
 */
#define RLDSP_NUM_CAPS   6


/*!
 * @brief DSP parameter 
 *
 * Refer comment of RLdspFormat for audio format.
 *
 * In rlib_dsp_control(), 
 * sets "fragment" to the given fragment size, and sets "numfrags" to two 
 * when executes ioctl ( ..., SNDCTL_DSP_SETFRAGMENT, ... ),
 *
 * By way of this, fragment size will be calculated as 
 * "numfrags"^^log_2{"fragment"}.
 * 
 * if you don't want to set fragment size, specify 0.
 */
typedef struct {
	int format;   /*!< audio format */
	int bits;     /*!< sampling data size */
	int rate;     /*!< sampling rate */
	int channel;  /*!< number of channnel */
	int blksize;  /*!< device driver buffer size */
	int fragment; /*!< device driver fragment size */
	int numfrags; /*!< number of fragment */
}
RLdspPrms;


/*!
 * @brief DSP audio format
 *
 * fmts represents if each format is  supported 
 * ( 1 when supported, 0 if not supported )
 *
 *   fmts[0] AFMT_MU_LAW<br>
 *   fmts[1] AFMT_A_LAW<br>
 *   fmts[2] AFMT_IMA_ADPCM<br>
 *   fmts[3] AFMT_U8<br>
 *   fmts[4] AFMT_S16_LE<br>
 *   fmts[5] AFMT_S16_BE<br>
 *   fmts[6] AFMT_S8<br>
 *   fmts[7] AFMT_U16_LE<br>
 *   fmts[8] AFMT_U16_BE<br>
 *   fmts[9] AFMT_MPEG<br>
 *
 */
typedef struct {
    int fmts[ RLDSP_NUM_FORMAT ];
}
RLdspFormat;


/*!
 * @brief DSP capabilities 
 *
 * revision means revision level.
 *
 * caps represents if each capability is supported.
 * ( 1 when supported, 0 if not supported )
 *
 * ########################################################<br>
 * # "multi" and "bind" is not supported under my environment<br>
 * ########################################################<br>
 *
 * caps[0] duplex       full duplex is supported<br>
 * caps[1] realtime     get precise position using SNDCTL_DSP_GET[IO]PTR<br>
 * caps[2] batch        use local strage for recording / playing<br>
 * caps[3] coprocessor  soundcard has real dsp, not codec<br>
 * caps[4] trigger      can use trigger<br>
 * caps[5] mmap         can use hardware buffer for recording / playing<br>
 * caps[6] multi        multiple open is supported<br>
 * caps[7] bind         can bind several channels ( front, rear, center )<br>
 *
 */
typedef struct {
    int revision;
    int caps[ RLDSP_NUM_CAPS ];
}
RLdspCaps;


int rlib_dsp_query_current( int fd, RLdspPrms* prms, int* err );
int rlib_dsp_control( int fd, RLdspPrms* prms, int* err );
int rlib_dsp_query_format( int fd, RLdspFormat* fmts, int* err );
int rlib_dsp_query_caps( int fd, RLdspCaps* caps, int* err );


#endif /* ___RLDSP_H_INCLUDED_ */

/* End of RLdsp.h */
