/*
 * Copyright (c) 2005-2007 Ryo Ichinose
 *
 * This file is part of Space Maze.
 * 
 * Space Maze is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; version 2 of the License.
 * 
 * Space Maze is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with Space Maze; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

/*!
 * @file
 *
 * Glue code to bind RLgt.h and Xlib functions 
 * which cannot be bound by "#define" keyword.
 *
 * $Id: RLgtx.c,v 1.2 2007/11/04 11:41:12 ryo Exp $
 */

#include <X11/Xlib.h>
#include <X11/keysym.h>

#include "RLerrno.h"
#include "RLgt.h"
#include "RLx.h"


typedef struct
{
	RLgtKey key;
	KeySym  keysym;
}
RLgtxKey;

typedef struct
{
	RLgtMouse mouse;
	unsigned  int button;
}
RLgtxMouse;

typedef struct
{
	RLgtEventType type;
	int xtype;
}
RLgtxEvent;

static RLgtxKey key_table[] =
{
	{ RL_GT_KEY_UP,     XK_Up     },
	{ RL_GT_KEY_DOWN,   XK_Down   },
	{ RL_GT_KEY_LEFT,   XK_Left   },
	{ RL_GT_KEY_RIGHT,  XK_Right  },
	{ RL_GT_KEY_ESC,    XK_Escape },
	{ RL_GT_KEY_RETURN, XK_Return },
	{ RL_GT_KEY_SPACE,  XK_space  },
	{ RL_GT_KEY_D,      XK_D      },
	{ RL_GT_KEY_F,      XK_F      },
	{ RL_GT_KEY_H,      XK_H      },
	{ RL_GT_KEY_M,      XK_M      },
	{ RL_GT_KEY_N,      XK_N      },
	{ RL_GT_KEY_O,      XK_O      },
	{ RL_GT_KEY_P,      XK_P      },
	{ RL_GT_KEY_S,      XK_S      },
	{ RL_GT_KEY_d,      XK_d      },
	{ RL_GT_KEY_f,      XK_f      },
	{ RL_GT_KEY_h,      XK_h      },
	{ RL_GT_KEY_m,      XK_m      },
	{ RL_GT_KEY_n,      XK_n      },
	{ RL_GT_KEY_o,      XK_o      },
	{ RL_GT_KEY_p,      XK_p      },
	{ RL_GT_KEY_s,      XK_s      },
	{ RL_GT_KEY_GUARD,  XK_p      } /* KeySym "XK_p" is dummy, will not to be used. */
};

static RLgtxMouse mouse_table[] =
{
	{ RL_GT_MOUSE_LEFT,   Button1 },
	{ RL_GT_MOUSE_MIDDLE, Button2 },
	{ RL_GT_MOUSE_RIGHT,  Button3 },
	{ RL_GT_MOUSE_GUARD,  Button3 } /* button "Button3" is dummy, not to be used. */
};

static RLgtxEvent event_table[] =
{
	{ RL_GT_KEY_PRESSED,    KeyPress      },
	{ RL_GT_KEY_RELEASED,   KeyRelease    },
	{ RL_GT_MOUSE_MOVED,    MotionNotify  },
	{ RL_GT_MOUSE_PRESSED,  ButtonPress   },
	{ RL_GT_MOUSE_RELEASED, ButtonRelease },
	{ RL_GT_EVENT_GUARD,    ButtonRelease } /* ButtonRelease is dummy, not to be used. */
};


/*!
 * Checks if the specified key is pressed at a certain frame.
 * Returns 1 when the specified key is pressed, returns 0 when not. 
 * Also returns 0 when app is NULL or key ( RLgtKey type ) 
 * is not bound to a certain key value on the platform.
 *
 * @param app [in] GUI toolkit context. Must not be NULL.
 * @param key [in] Key to check.
 *
 * @retval 0 Specified key is not pressed.
 * @retval 1 Specified key is pressed.
 */
int
rlib_gt_key_pressed( RLgtApp* app, RLgtKey key )
{
	int i = 0;

	if ( app == NULL ) {
		return 0;
	}

	for ( ;; i++ ) {
		if ( key_table[i].key == RL_GT_KEY_GUARD ) {
			return 0;
		}
		
		if ( key == key_table[i].key ) {
			return rlib_x_key_pressed( app, key_table[i].keysym );
		}
	}
}


/*!
 * Checks if the specified key is released at a certain frame.
 * Returns 1 when the specified key is released, returns 0 when not. 
 * Also returns 0 when app is NULL or key ( RLgtKey type ) 
 * is not bound to a certain key value on the platform.
 *
 * @param app [in] GUI toolkit context. Must not be NULL.
 * @param key [in] Key to check.
 *
 * @retval 0 Specified key is not released.
 * @retval 1 Specified key is released.
 */
int
rlib_gt_key_released( RLgtApp* app, RLgtKey key )
{
	int i = 0;

	if ( app == NULL ) {
		return 0;
	}

	for ( ;; i++ ) {
		if ( key_table[i].key == RL_GT_KEY_GUARD ) {
			return 0;
		}
		
		if ( key == key_table[i].key ) {
			return rlib_x_key_released( app, key_table[i].keysym );
		}
	}
}


/*!
 * Checks if the specified mouse button is pressed at a certain frame.
 * Returns 1 when the specified mouse button is pressed, returns 0 when not. 
 * Also returns 0 when app is NULL / mouse button ( RLgtMouse type ) 
 * is not bound to a certain button on the platform.
 *
 * If x or y is not NULL, pressed position will be returned.
 *
 * @param app   [in]  GUI toolkit context. Must not be NULL.
 * @param mouse [in]  Mouse button to check.
 * @param x     [out] X position where mouse button is pressed when not NULL.
 * @param y     [out] Y position where mouse button is pressed when not NULL.
 *
 * @retval 0 Specified mouse button is not pressed.
 * @retval 1 Specified mouse button is pressed.
 */
int
rlib_gt_mouse_pressed( RLgtApp* app, RLgtMouse mouse, int* x, int* y )
{
	int i = 0;

	if ( app == NULL ) {
		return 0;
	}

	for ( ;; i++ ) {
		if ( mouse_table[i].mouse == RL_GT_MOUSE_GUARD ) {
			return 0;
		}
		
		if ( mouse == mouse_table[i].mouse ) {
			return rlib_x_mouse_pressed( app, mouse_table[i].button, x, y );
		}
	}
}


/*!
 * Checks if the specified mouse button is pressed at a certain frame.
 * Returns 1 when the specified mouse button is released, returns 0 when not. 
 * Also returns 0 when app is NULL / mouse button ( RLgtMouse type ) 
 * is not bound to a certain button on the platform.
 *
 * If x or y is not NULL, released position will be returned.
 *
 * @param app   [in]  GUI toolkit context. Must not be NULL.
 * @param mouse [in]  Mouse button to check.
 * @param x     [out] X position where mouse button is released when not NULL.
 * @param y     [out] Y position where mouse button is released when not NULL.
 *
 * @retval 0 Specified mouse button is not released.
 * @retval 1 Specified mouse button is released.
 */
int
rlib_gt_mouse_released( RLgtApp* app, RLgtMouse mouse, int* x, int* y )
{
	int i = 0;

	if ( app == NULL ) {
		return 0;
	}

	for ( ;; i++ ) {
		if ( mouse_table[i].mouse == RL_GT_MOUSE_GUARD ) {
			return 0;
		}
		
		if ( mouse == mouse_table[i].mouse ) {
			return rlib_x_mouse_released( app, mouse_table[i].button, x, y );
		}
	}
}


/*!
 * Adds event handler that will be called when specified event occures.
 *
 * @param app  [in] Application context of gui toolkit. Must not be NULL.
 * @param type [in] Event type. Specify one of the following value.
 *
 *   - RL_GT_KEY_PRESSED
 *   - RL_GT_KEY_RELEASED
 *   - RL_GT_MOUSE_MOVED
 *   - RL_GT_MOUSE_PRESSED
 *
 * @param callback [in] Call back function called when specified event occures.
 * @param data     [in] This pointer will be passed to call back function.
 *
 * @retval RLIB_EFAULT app is NULL or callback is NULL, or type is invalid.
 * @retval RLIB_ENOSPC Maximum number of callbacks is already registered.
 */
int
rlib_gt_add_event_handler( RLgtApp* app, RLgtEventType type, int ( *callback )( void* ), void* data )
{
	int i = 0;

	if ( app == NULL || callback == NULL ) {
		return RLIB_EFAULT;
	}

	for ( ;; i++ ) {
		if ( event_table[i].type == RL_GT_EVENT_GUARD ) {
			return RLIB_EFAULT;
		}

		if ( type == event_table[i].type ) {
			return rlib_x_add_event_handler( app, event_table[i].xtype, callback, data );
		}
	}
}


/*!
 * Sets foreground color which will be used to draws something on screen.
 * Specify RGB value as r (RED), g (GREEN), b (BLUE).
 * Each value must be specified between 0 and RL_GT_MAX_RGB.
 * When app is NULL, do nothing.
 */
void
rlib_gt_set_foreground_color( RLgtApp* app, unsigned int r, unsigned int g, unsigned int b )
{
	int _r = ( r > RL_GT_MAX_RGB ) ? RL_GT_MAX_RGB : r;
	int _g = ( g > RL_GT_MAX_RGB ) ? RL_GT_MAX_RGB : g;
	int _b = ( b > RL_GT_MAX_RGB ) ? RL_GT_MAX_RGB : b;

	rlib_x_set_foreground_color( app, RL_GT_MAX_RGB, _r, _g, _b );
}


/*!
 * Sets background color which will be used to draws something on screen.
 * Specify RGB value as r (RED), g (GREEN), b (BLUE).
 * Each value must be specified between 0 and RL_GT_MAX_RGB.
 * When app is NULL, do nothing.
 */
void
rlib_gt_set_background_color( RLgtApp* app, unsigned int r, unsigned int g, unsigned int b )
{
	int _r = ( r > RL_GT_MAX_RGB ) ? RL_GT_MAX_RGB : r;
	int _g = ( g > RL_GT_MAX_RGB ) ? RL_GT_MAX_RGB : g;
	int _b = ( b > RL_GT_MAX_RGB ) ? RL_GT_MAX_RGB : b;

	rlib_x_set_background_color( app, RL_GT_MAX_RGB, _r, _g, _b );
}


/*!
 * Fill polygon.
 *
 * @param app     [in] GUI toolkit context. Must not be NULL.
 * @param points  [in] Point data is stored. All points must be specified
 *                     with absolute position correspond to the window,
 *                     not relative position to adjacent points.
 *
 * @param npoints [in] Number of points.
 * @param mode    [in] Specify one of the following. 
 *                     RL_GT_POLYGON_CONVEX | RL_GT_POLYGON_NONCONVEX | RL_GT_POLYGON_COMPLEX.
 *                     If any other value is specified, RL_GT_POLYGON_CONVEX will be used.
 */
void
rlib_gt_fill_polygon( RLgtApp* app, Point* points, int npoints, RLgtPolygonMode mode )
{
	int _mode;

	switch ( mode ) {
		case RL_GT_POLYGON_NONCONVEX :
			_mode = Nonconvex;
			break;

		case RL_GT_POLYGON_COMPLEX :
			_mode = Complex;
			break;

		default :
			_mode = Convex;
			break;
	}

	rlib_x_fill_polygon( app, points, npoints, _mode );
}


/*!
 * Loads the specified image file and saves it into application context.
 * When successfully processed, returns index to refer loaded image.
 * Use this index when you call rlib_gt_draw_image() / rlib_x_get_image_size().
 *
 * @param app      [in] GUI toolkit context. Must not be NULL.
 * @param filename [in] Image file name ( full path ). Imlib supported images can be loaded.
 * @param use_mask [in] Specify RL_GT_USE_MASK when you want to create mask of the image.
 *                      Else, specify RL_GT_NO_MASK. 
 *                      RL_GT_USE_MASK will be used as default.
 *
 * @retval RLIB_EFAULT app is NULL or filename is NULL.
 * @retval RLIB_ENOSPC Maximum number of pixmaps is already loaded.
 */
int
rlib_gt_load_image( RLgtApp* app, char* filename, int use_mask  )
{
	int _use_mask = ( use_mask == RL_GT_NO_MASK ) ? RL_X_NO_MASK : RL_X_USE_MASK;

	if ( app == NULL || filename == NULL ) {
		return RLIB_EFAULT;
	}

	return rlib_x_load_image( app, filename, _use_mask );
}


/* End of RLgtx.c */


