/*
 * Copyright (c) 2006 Ryo Ichinose
 *
 * This file is part of Space Maze.
 * 
 * Space Maze is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; version 2 of the License.
 * 
 * Space Maze is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with Space Maze; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

/*!
 * @file
 * @brief Sound library using OSS ( Open Sound System ).
 *
 * $Id: RLoss.c,v 1.3 2006/10/09 09:10:16 ryo Exp $
 */

#include <fcntl.h>
#include <limits.h>
#include <linux/soundcard.h>
#include <pthread.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include "RLerrno.h"
#include "RLdsp.h"
#include "RLlist.h"
#include "RLlogger.h"
#include "RLoss.h"
#include "RLwave.h"


#define RLOSS_LOGBUF_SIZE   512
#define RLOSS_SAMPLING_BITS 16
#define RLOSS_SAMPLING_RATE 11025
#define RLOSS_CHANNEL       1
#define RLOSS_FRAGMENT_SIZE 1024
#define RLOSS_FRAGMENT_NUM  4


static void* rlib_oss_playing_thread( void* param );
static int   rlib_oss_mix( short* dst, short* src, int size );
static void  rlib_oss_cleanup_playlist( RLossContext* context );


/*!
 * Initializes sound library and starts playing thread. 
 * Returns 0 when successfully processed.
 * Other value means initialization failed.
 *
 * @param context     [i/o] Sound library context. Must not be NULL.
 * @param device_name [in]  Sound device name. Must not be NULL.
 */
int
rlib_oss_initialize( RLossContext* context, char* device_name )
{
	RLdspPrms dsp_params = { 
		AFMT_S16_LE, 
		RLOSS_SAMPLING_BITS,
		RLOSS_SAMPLING_RATE, 
		RLOSS_CHANNEL,
		0, 
		RLOSS_FRAGMENT_SIZE,
		RLOSS_FRAGMENT_NUM
	};

	pthread_attr_t attr;
	char logbuf[ RLOSS_LOGBUF_SIZE ];
	int err;
	int ret;

	if ( context == NULL || device_name == NULL ) {
		return RLIB_ERR;
	}

	memset( context, 0, sizeof( RLossContext ) );
	rlib_list_init( &( context->play_list ) );

	if ( ( context->fd = open( device_name, O_WRONLY ) ) < 0 ) {
		snprintf( logbuf, sizeof( logbuf ), "Failed to open sound device %s.", device_name );
		rlib_logger_warn( logbuf );
		return RLIB_ERR;
	}

	if ( rlib_dsp_query_current( context->fd, &( context->device_prev_params ), &err ) ) {
		snprintf( logbuf, sizeof( logbuf ), "Failed to get sound device parameters, errno = %d", err );
		rlib_logger_warn( logbuf );
		goto err_end;
	}

	if ( rlib_dsp_control( context->fd, &dsp_params, &err ) ) {
		snprintf( logbuf, sizeof( logbuf ), "Failed to set sound device parameters, errno = %d", err );
		rlib_logger_warn( logbuf );
		goto err_end;
	}

	pthread_mutex_init( &( context->mutex ), NULL );
	pthread_attr_init( &attr );
	pthread_attr_setdetachstate( &attr, PTHREAD_CREATE_JOINABLE );

	ret = pthread_create( &( context->thread_id ), &attr, rlib_oss_playing_thread, ( void* )context );
	pthread_attr_destroy( &attr );

	if ( ret ) {
		snprintf( logbuf, sizeof( logbuf ), "Failed to create thread, result code = %d", ret );
		rlib_logger_warn( logbuf );
		goto err_end;
	}

	context->initialized = 1;
	return 0;

 err_end:
	close( context->fd );
	context->fd = -1;
	context->initialized = 0;
	return RLIB_ERR;
}


/*!
 * Loads specified wave file and holds its pcm data within sound library context.
 * Only the wave file with following format can be load.
 * When successfully processed, returns id to refer loaded wave sound.
 * Use this id when you call rlib_oss_play().
 *
 * - Sampling bits : 16 bit
 * - Sampling rate : 11025Hz
 * - Channel       : monoral
 *
 * @param context   [i/o] Sound library context.
 * @param wave_name [in]  Wave file name to load.
 *
 * @retval RLIB_ENOSPC  Maximum number of wave files are already loaded.
 * @retval RLIB_EFAULT  context or wave_name is NULL, or context is not initialized.
 * @retval RLIB_ERR     Failed to load specified wave file.
 */
int
rlib_oss_load_wave( RLossContext* context, char* wave_name )
{
	RLWAVE wave;
	char logbuf[ RLOSS_LOGBUF_SIZE ];
	int id;
	int err;
	int ret;

	if ( context == NULL || ! context->initialized || wave_name == NULL ) {
		return RLIB_EFAULT;
	}

	if ( context->num_of_waves > RL_OSS_MAX_WAVE_NUM ) {
		return RLIB_ENOSPC;
	}

	if ( ( ret = rlib_wave_open( &wave, wave_name, RLWAVE_READ, &err ) ) ) {
		snprintf( logbuf, sizeof( logbuf ), "Failed to load wave file %s, "
			  "ret = %d, errno = %d", wave_name, ret, err );
		rlib_logger_warn( logbuf );
		return RLIB_ERR;
	}

	if ( ( ret = rlib_wave_read_header( &wave, &err ) ) ) {
		snprintf( logbuf, sizeof( logbuf ), "Failed to read wave file %s, "
			  "ret = %d, errno = %d", wave_name, ret, err );
		rlib_logger_warn( logbuf );
		return RLIB_ERR;
	}

	if ( ! wave.is_pcm || wave.rate != RLOSS_SAMPLING_RATE || 
	     wave.channel != RLOSS_CHANNEL || wave.bits != RLOSS_SAMPLING_BITS ) {
		snprintf( logbuf, sizeof( logbuf ), 
			  "Wave file must be PCM / 16bits / 11025Hz / monoral : %s", wave_name );
		rlib_logger_warn( logbuf );
		return RLIB_ERR;
	}

	pthread_mutex_lock( &( context->mutex ) );

	id = context->num_of_waves ++;
	memset( &( context->waves[ id ] ), 0, sizeof( RLossWaveData ) );
	context->waves[ id ].length = wave.len;
	context->waves[ id ].data   = ( short* )malloc( wave.len );
	fseek( wave.fp, wave.offset, SEEK_SET );
	fread( ( void* )( context->waves[ id ].data ), 1, wave.len, wave.fp );
	rlib_wave_close( &wave );

	pthread_mutex_unlock( &( context->mutex ) );
	return id;
}


/*!
 * Adds id specified wave sound to play list.
 * When you want to play loop, specify 1 to "loop".
 * When successfully processed, returns id to refer playing sound.
 * Use this id when you call rlib_oss_stop().
 * Note that returned id might be different from the one you specified to this function.
 *
 * @param context Sound library context. Must not be NULL.
 * @param id      Sound id returned when you call rlib_oss_load_wave().
 * @param loop    Specify 1 when you want to play sound loop, otherwise specify 0.
 *
 * @retval RLIB_EFAULT context is NULL, or context is not initialized.
 * @retval RLIB_ENOENT Wave sound with specified id is not loaded.
 */
int  
rlib_oss_play( RLossContext* context, int id, int loop )
{
	RLossPlayItem item;

	if ( context == NULL || ! context->initialized ) {
		return RLIB_EFAULT;
	}

	pthread_mutex_lock( &( context->mutex ) );

	if ( id >= context->num_of_waves ) {
		pthread_mutex_unlock( &( context->mutex ) );
		return RLIB_ENOENT;
	}

	item.id         = context->next_play_id ++;
	item.wave_index = id;
	item.loop       = loop;
	item.current    = 0;

	rlib_list_insert( &( context->play_list ), -1, ( unsigned char* )&item, sizeof( RLossPlayItem ) );
	pthread_mutex_unlock( &( context->mutex ) );

	return 0;
}


/*!
 * Stop id specified playing sound. This id must be the one rlib_oss_play() returned.
 * @param context [i/o] Sound library context. Must not be NULL.
 * @param id      [in]  Sound id returned when you call rlib_oss_play().
 */
void
rlib_oss_stop( RLossContext* context, int id )
{
	int i;

	if ( context == NULL || ! context->initialized ) {
		return;
	}

	pthread_mutex_lock( &( context->mutex ) );

	for ( i = 0; i < rlib_list_size( &( context->play_list ) ); i ++ ) {
		RLlistData* list_data = rlib_list_get( &( context->play_list ), i );
		RLossPlayItem* item = ( RLossPlayItem* )( list_data->buf );

		if ( item->id == id ) {
			rlib_list_remove( &( context->play_list ), i );
			break;
		}
	}

	pthread_mutex_unlock( &( context->mutex ) );
}


/*!
 * Stops all the playing sound.
 */
void
rlib_oss_stop_all( RLossContext* context )
{
	if ( context == NULL || ! context->initialized ) {
		return;
	}

	pthread_mutex_lock( &( context->mutex ) );
	rlib_list_clear( &( context->play_list ) );
	pthread_mutex_unlock( &( context->mutex ) );
}


/*!
 * Stops playing thread and shutdown sound library.
 */
void
rlib_oss_shutdown( RLossContext* context )
{
	int i;

	if ( context == NULL || ! context->initialized ) {
		return;
	}

	pthread_mutex_lock( &( context->mutex ) );
	context->stopped ++;
	pthread_mutex_unlock( &( context->mutex ) );
 	pthread_join( context->thread_id, NULL );

	rlib_dsp_control( context->fd, &( context->device_prev_params ), NULL );
	close( context->fd );

	for ( i = 0; i < context->num_of_waves; i ++ ) {
		free( context->waves[i].data );
	}

	pthread_mutex_destroy( &( context->mutex ) );
}


/*!
 * Worker thread that plays sound.
 */
static void*
rlib_oss_playing_thread( void* param )
{
	RLossContext* context = ( RLossContext* )param;
	int mixed;
	int i;

	if ( context == NULL ) {
		pthread_exit( NULL );
	}

	while ( 1 ) {
		if ( ! context->initialized ) {
			continue;
		}

		pthread_mutex_lock( &( context->mutex ) );
		memset( ( void* )( context->buffer ), 0, RL_OSS_BUFSIZE );

		for ( i = 0; i < rlib_list_size( &( context->play_list ) ); i ++ ) {
			RLlistData* list_data = rlib_list_get( &( context->play_list ), i );
			RLossPlayItem* item = ( RLossPlayItem* )( list_data->buf );

			short* wave_data = context->waves[ item->wave_index ].data;
			int wave_len = context->waves[ item->wave_index ].length;
			mixed = 0;

			while ( 1 ) {
				int src_left = wave_len - item->current;
				int dst_left = RL_OSS_BUFSIZE - mixed;
				int mix_len  = src_left > dst_left ? dst_left : src_left;

				rlib_oss_mix( context->buffer + mixed / sizeof( short ), 
					      wave_data + item->current / sizeof( short ), 
					      mix_len );

				mixed += mix_len;
				item->current += mix_len;

				if ( item->loop && item->current >= wave_len ) {
					item->current -= wave_len;
				}

				if ( mixed >= RL_OSS_BUFSIZE || item->current >= wave_len ) {
					break;
				}
			}
		}

		rlib_oss_cleanup_playlist( context );
		pthread_mutex_unlock( &( context->mutex ) );
		write( context->fd, ( void* )( context->buffer ), RL_OSS_BUFSIZE );

		if ( context->stopped ) {
			break;
		}
	}

	pthread_exit( NULL );
}


#define MIN( a, b ) ( ( (a) < (b) ) ? (a) : (b) )
#define MAX( a, b ) ( ( (a) > (b) ) ? (a) : (b) )


static int 
rlib_oss_mix( short* dst, short* src, int size )
{
	int i;

	if ( dst == NULL || src == NULL ) {
		return RLIB_EFAULT;
	}

	for ( i = 0; i < size / sizeof( short ); i ++ ) {
		if ( src[i] < 0 ) {
			dst[i] = ( short )MAX( (int)dst[i] + (int)src[i], (int)SHRT_MIN );
		}
		else {
			dst[i] = ( short )MIN( (int)dst[i] + (int)src[i], (int)SHRT_MAX );
		}
	}

	return 0;
}


static void
rlib_oss_cleanup_playlist( RLossContext* context )
{
	RLlistData* list_data;
	RLossPlayItem* item;
	int i = 0;

	while ( 1 ) {
		if ( i == rlib_list_size( &( context->play_list ) ) ) {
			break;
		}

		list_data = rlib_list_get( &( context->play_list ), i );
		item = ( RLossPlayItem* )( list_data->buf );

		if ( ! item->loop && item->current >= context->waves[ item->wave_index ].length ) {
			rlib_list_remove( &( context->play_list ), i );
		}
		else {
			i ++;
		}
	}
}


/* End of RLoss.c */
