/*
 * Copyright (c) 2005,2006 Ryo Ichinose
 *
 * This file is part of Space Maze.
 * 
 * Space Maze is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; version 2 of the License.
 * 
 * Space Maze is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with Space Maze; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

/*!
 * @file
 * @brief Header file of X application tools and related functions.
 *
 * $Id: RLx.h,v 1.1.1.1 2006/07/22 18:07:56 ryo Exp $
 */

#ifndef ___RL_X_H_INCLUDED_
#define ___RL_X_H_INCLUDED_

#include <X11/Xlib.h>
#include <Imlib.h>
#include "RLlist.h"


/*!
 * @brief Max pixmap num that can hold within RLXApp context.
 */
#define RL_X_MAX_PIXMAP_NUM 128


/*!
 * @brief Specify this value when you call rlib_x_load_xpm() / rlib_x_load_png()
 * if you want to create mask pixmap.
 */
#define RL_X_USE_MASK  1


/*!
 * @brief Specify this value when you call rlib_x_load_xpm() / rlib_x_load_png()
 * if you don't want to create mask pixmap.
 */
#define RL_X_NO_MASK   0


/*!
 * @brief Data structure to manage pixmap and mask of it.
 */
typedef struct 
{
	Pixmap body;
	Pixmap mask;
	int use_mask;
	int width;
	int height;
}
RLXPixmap;


/*!
 * @brief Data structure to register event handler.
 */
typedef struct
{
	int   type;
	int   ( *callback )( void* );
	void* data;
}
RLXEventHandler;


/*!
 * @brief X application context.
 */
typedef struct _RLXApp
{
	/* Basic X resources. */
	Display*   display;
	Window     root;
	int        screen;
	Colormap   cmap;
	GC         gc;
	GC         gc_debug;
	GC         gc_pixmap;
	unsigned   long white;
	unsigned   long black;

	/* Window parameters. */
	Drawable   frontbuffer;
	Drawable   backbuffer;
	int        x;
	int        y;
	unsigned   int width;
	unsigned   int height;
	unsigned   int display_width;
	unsigned   int display_height;

	/* Event loop related things. */
	Atom       atom_delete_window;
	int        do_loop;
	int        drop_frame;
	int        event_renewed;
	int        exitcode;
	int        fps;
	struct     timeval previous;
	XEvent     event;
	RLlist     event_handlers;
	void       ( *update_func )( struct _RLXApp* );
	void       ( *draw_func )( struct _RLXApp* );

	/* Image related things. */
	RLXPixmap  pixmaps[ RL_X_MAX_PIXMAP_NUM ];
	ImlibData* imlib_data;
	int        num_of_pixmaps;

	/* Any other things. */
	int        autorepeat_previous;
	int        debug;
	void       ( *cleanup_func )( void* );
	void*      data_for_cleanup;
}
RLXApp;


int  rlib_x_initialize( RLXApp* app, char* title, int x, int y, int width, int height );
int  rlib_x_get_width( RLXApp* app );
int  rlib_x_get_height( RLXApp* app );
void rlib_x_enable_dropframe( RLXApp* app );
void rlib_x_disable_dropframe( RLXApp* app );
int  rlib_x_is_dropframe_enabled( RLXApp* app );
void rlib_x_set_fps( RLXApp* app, int fps );
void rlib_x_set_update_func( RLXApp* app, void ( *func )( RLXApp* ) );
void rlib_x_set_draw_func( RLXApp* app, void ( *func )( RLXApp* ) );
void rlib_x_set_cleanup_func( RLXApp* app, void ( *func )( void* ), void* data );
void rlib_x_autorepeat_on( RLXApp* app );
void rlib_x_autorepeat_off( RLXApp* app );
void rlib_x_autorepeat_set( RLXApp* app, int autorepeat );
int  rlib_x_get_autorepeat_previous( RLXApp* app );
int  rlib_x_get_autorepeat( RLXApp* app );
void rlib_x_start( RLXApp* app );
void rlib_x_stop( RLXApp* app, int exitcode );
void rlib_x_cleanup( RLXApp* app );
int  rlib_x_get_exitcode( RLXApp* app, int* exitcode );
int  rlib_x_get_drawable( RLXApp* app, Drawable* drawable );
long rlib_x_sync( RLXApp* app );
void rlib_x_debug_on( RLXApp* app );
void rlib_x_debug_off( RLXApp* app );
int  rlib_x_is_debug_on( RLXApp* app );
XEvent* rlib_x_get_event( RLXApp* app );
int  rlib_x_key_pressed( RLXApp* app, KeySym keysym );
int  rlib_x_key_released( RLXApp* app, KeySym keysym );
int  rlib_x_mouse_pressed( RLXApp* app, unsigned int button, int* x, int* y );
int  rlib_x_mouse_released( RLXApp* app, unsigned int button, int* x, int* y );
int  rlib_x_mouse_moved( RLXApp* app, int* x, int* y );
void rlib_x_mouse_query_position( RLXApp* app, int* x, int* y );
int  rlib_x_add_event_handler( RLXApp* app, int type, int ( *callback )( void* ), void* data );
void rlib_x_clear_event_handler( RLXApp* app );
int  rlib_x_process_event( RLXApp* app );

int  rlib_x_draw_pixmap( RLXApp* app, int index, 
      int src_x, int src_y, int width, int height, int dst_x, int dst_y );

int  rlib_x_load_image( RLXApp* app, char* filename, int use_mask );
int  rlib_x_get_pixmap_size( RLXApp* app, int index, int* width, int* height );

void rlib_x_set_foreground_color( 
	RLXApp* app, int scale, unsigned int r, unsigned int g, unsigned int b );

void rlib_x_set_background_color( 
	RLXApp* app, int scale, unsigned int r, unsigned int g, unsigned int b );

void rlib_x_draw_point( RLXApp* app, int x, int y );
void rlib_x_draw_line( RLXApp* app, int src_x, int src_y, int dst_x, int dst_y );
void rlib_x_draw_lines( RLXApp* app, XPoint* points, int npoints );
void rlib_x_fill_polygon( RLXApp* app, XPoint* points, int npoints, int mode );
void rlib_x_draw_rectangle( RLXApp* app, int x, int y, int width, int height );
void rlib_x_fill_rectangle( RLXApp* app, int x, int y, int width, int height );
void rlib_x_draw_arc( RLXApp* app, int x, int y, int width, int height, int angle1, int angle2 );
void rlib_x_fill_arc( RLXApp* app, int x, int y, int width, int height, int angle1, int angle2 );


#endif /* ___RL_X_H_INCLUDED_ */

/* End of RLx.h */
