/*
 * Copyright (c) 2005-2007 Ryo Ichinose
 *
 * This file is part of Space Maze.
 * 
 * Space Maze is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; version 2 of the License.
 * 
 * Space Maze is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with Space Maze; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

/*!
 * @file
 * @brief X application tools ( core functions ).
 *
 * $Id: RLxcore.c,v 1.2 2007/11/04 11:41:12 ryo Exp $
 */

#include <stdio.h>
#include <string.h>
#include <sys/time.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>

#include "RLerrno.h"
#include "RLlogger.h"
#include "RLtimer.h"
#include "RLx.h"


static void rlib_x_init_context( RLXApp* app, int x, int y, int width, int height );
static void rlib_x_debug_draw( RLXApp* app, long elapsed_time, int frame_count );
static void rlib_x_flip( RLXApp* app );
static long rlib_x_get_elapsed_time( RLXApp* app, struct timeval* current_return );


#define RL_X_DELETE_WINDOW     "WM_DELETE_WINDOW"
#define RL_X_DEFAULT_WINDOW_W  640
#define RL_X_DEFAULT_WINDOW_H  480
#define RL_X_DEFAULT_WINDOW_X  100
#define RL_X_DEFAULT_WINDOW_Y  100
#define RL_X_DEFAULT_FPS       60


/*!
 * Initializes application context of gui toolkit.
 * This function creates window with specified title, position and
 * size ( specified by width and height ).
 *
 * If you don't want to specify title, x / y, width / height and
 * leave those values as default, specify each values as
 * NULL, -1 / -1, -1 / -1.
 *
 * Created window cannot be resized. Following event mask will 
 * be specified.
 *
 * <ul>
 *   <li>KeyPressMask
 *   <li>KeyReleaseMask
 *   <li>ButtonPressMask
 *   <li>ButtonReleaseMask
 *   <li>PointerMotionMask
 *   <li>PointerMotionHintMask
 *   <li>FocusChangeMask
 * </ul>
 *
 * @param app         [i/o] GUI toolkit context. Must not be NULL.
 * @param title       [in]  Window title.
 * @param x           [in]  x position. Specify -1 to use default value.
 * @param y           [in]  y position. Specify -1 to use default value.
 * @param width       [in]  Window width. Specify -1 to use default value.
 * @param height      [in]  Window height. Specify -1 to use default value.
 *
 * @retval 0           Successfully processed.
 * @retval RLIB_EFAULT Failed to connect to X server, or app is NULL.
 */
int
rlib_x_initialize( RLXApp* app, char* title, int x, int y, int width, int height )
{
	if ( app == NULL ) {
		return RLIB_EFAULT;
	}

	rlib_x_init_context( app, x, y, width, height );

	if ( ( app->display = XOpenDisplay( NULL ) ) == NULL ) {
		rlib_logger_fatal( "Failed to connect to X server." );
		return RLIB_EFAULT;
	}

	app->root           = DefaultRootWindow( app->display );
	app->screen         = DefaultScreen( app->display );
	app->cmap           = DefaultColormap( app->display, app->screen );
	app->white          = WhitePixel( app->display, app->screen );
	app->black          = BlackPixel( app->display, app->screen );
	app->display_width  = DisplayWidth( app->display, app->screen );
	app->display_height = DisplayHeight( app->display, app->screen );

	app->autorepeat_previous = rlib_x_get_autorepeat( app );
	app->imlib_data = Imlib_init( app->display );

	app->frontbuffer = 
		XCreateSimpleWindow( 
			app->display, 
			app->root,
			app->x, 
			app->y, 
			app->width, 
			app->height, 
			0,
			app->black, 
			app->white );

	if ( title != NULL ) {
		XStoreName( app->display, app->frontbuffer, title );
	}

	app->backbuffer = 
		XCreatePixmap(
			app->display,
			app->root,
			app->width,
			app->height,
			DefaultDepth( app->display, app->screen )
			);

	{
		unsigned long mask = ( 0L | GCGraphicsExposures );
		XGCValues values;
		values.graphics_exposures = False;

		app->gc        = XCreateGC( app->display, app->frontbuffer, mask, &values );
		app->gc_debug  = XCreateGC( app->display, app->frontbuffer, mask, &values );
		app->gc_pixmap = XCreateGC( app->display, app->frontbuffer, mask, &values );

		XSetForeground( app->display, app->gc_debug, app->black );
		XSetBackground( app->display, app->gc_debug, app->white );
	}

	/*
	 * Disable resize feature.
	 */
	{
		XSizeHints hints;

		hints.min_width  = app->width;
		hints.min_height = app->height;
		hints.max_width  = app->width;
		hints.max_height = app->height;
		hints.flags = PMinSize | PMaxSize;

		XSetWMNormalHints( app->display, app->frontbuffer, &hints );
	}

	/*
	 * Ask window manager to notify me when my window is going to be closed for some reason.
	 */
	app->atom_delete_window = XInternAtom( app->display, RL_X_DELETE_WINDOW, True );
	XSetWMProtocols( app->display, app->frontbuffer, &( app->atom_delete_window ), 1 );

	XSelectInput(
		app->display,
		app->frontbuffer,
		KeyPressMask      | 
		KeyReleaseMask    | 
		ButtonPressMask   | 
		ButtonReleaseMask | 
		PointerMotionMask | 
		PointerMotionHintMask |
		FocusChangeMask
		);

	return 0;
}


static void
rlib_x_init_context( RLXApp* app, int x, int y, int width, int height )
{
	memset( app, 0, sizeof( RLXApp ) );
	timerclear( &( app->previous ) );
	rlib_list_init( &( app->event_handlers ) );
	app->fps = RL_X_DEFAULT_FPS;

	app->x = ( x < 0 ) ? RL_X_DEFAULT_WINDOW_X : x;
	app->y = ( y < 0 ) ? RL_X_DEFAULT_WINDOW_Y : y;
	app->width  = ( width  < 0 ) ? RL_X_DEFAULT_WINDOW_W : width;
	app->height = ( height < 0 ) ? RL_X_DEFAULT_WINDOW_H : height;
}


/*!
 * Gets the width of the created window.
 * Returns 0 when app is NULL.
 */
int
rlib_x_get_width( RLXApp* app )
{
	if ( app == NULL ) {
		return 0;
	}

	return app->width;
}


/*!
 * Gets the height of the created window.
 * Returns 0 when app is NULL.
 */
int 
rlib_x_get_height( RLXApp* app )
{
	if ( app == NULL ) {
		return 0;
	}

	return app->height;
}


/*!
 * Enables "drop frame" function in event loop.
 */
void
rlib_x_enable_dropframe( RLXApp* app )
{
	if ( app != NULL ) {
		app->drop_frame = 1;
	}
}


/*!
 * Disables "drop frame" function in event loop.
 */
void
rlib_x_disable_dropframe( RLXApp* app )
{
	if ( app != NULL ) {
		app->drop_frame = 0;
	}
}


/*!
 * Checks if "drop frame" function is enabled.
 * Returns 1 when enabled, 0 when not.
 */
int
rlib_x_is_dropframe_enabled( RLXApp* app )
{
	if ( app == NULL ) {
		return 0;
	}

	return app->drop_frame;
}


/*!
 * Sets fps ( frames per second ) to the given value.
 * fps must be more than 0 and less than or equal to 
 * RL_X_DEFAULT_FPS.
 */
void
rlib_x_set_fps( RLXApp* app, int fps )
{
	if ( app != NULL && fps > 0 && fps <= RL_X_DEFAULT_FPS ) {
		app->fps = fps;
	}
}


/*!
 * Function set here will be called at every event loop before 
 * drawing the screen. If you want to do something at every event loop
 * right after reading X event ( if it exists ), specify it here.
 */
void
rlib_x_set_update_func( RLXApp* app, void ( *func )( RLXApp* ) )
{
	if ( app != NULL ) {
		app->update_func = func;
	}
}


/*!
 * Function set here will be called to draw the screen in event loop. 
 * When "drop frame" function is enabled and a certain event loop takes 
 * longer time than necessary time to realize required fps, specified 
 * drawing function will not be called at some frame ( only the update
 * function specified by rlib_x_set_update_func will be called ).
 */
void
rlib_x_set_draw_func( RLXApp* app, void ( *func )( RLXApp* ) )
{
	if ( app != NULL ) {
		app->draw_func = func;
	}
}


/*!
 * Function set here will be called when event loop is going to end,
 * and it will be called before internal cleanup function is called
 * ( this means before X server connection is closed ).
 * If you want to do something to clean up out of the X event loop
 * scope, specify it here. 
 *
 * "data" will be passed to the specified function when called.
 */
void
rlib_x_set_cleanup_func( RLXApp* app, void ( *func )( void* ), void* data )
{
	if ( app != NULL ) {
		app->cleanup_func = func;
		app->data_for_cleanup = data;
	}
}


/*!
 * Enables keyboard autorepeat function.
 */
void
rlib_x_autorepeat_on( RLXApp* app )
{
	if ( app != NULL ) {
		XAutoRepeatOn( app->display );
	}
}


/*!
 * Disables keyboard autorepeat function.
 */
void
rlib_x_autorepeat_off( RLXApp* app )
{
	if ( app != NULL ) {
		XAutoRepeatOff( app->display );
	}
}


/*!
 * Sets keyboard autorepeat function to the specified mode.
 * "autorepeat" must be AutoRepeatModeOn or AutoRepeatModeOff.
 */
void
rlib_x_autorepeat_set( RLXApp* app, int autorepeat )
{
	if ( app != NULL ) {
		if ( autorepeat == AutoRepeatModeOn || autorepeat == AutoRepeatModeOff ) {
			XKeyboardControl values;
			unsigned long value_mask;

			values.auto_repeat_mode = autorepeat;
			value_mask = ( 0 | KBAutoRepeatMode );

			XChangeKeyboardControl( app->display, value_mask, &values );
		}
	}
}


/*!
 * Gets keyboard autorepeat mode when rlib_x_initialize is called,
 * and returns AutoRepeatModeOn or AutoRepeatModeOff.
 * Returns RLIB_EFAULT when app is NULL.
 */
int
rlib_x_get_autorepeat_previous( RLXApp* app )
{
	if ( app == NULL ) {
		return RLIB_EFAULT;
	}

	return app->autorepeat_previous;
}


/*!
 * Gets current keyboard autorepeat mode.
 * Returns AutoRepeatModeOn or AutoRepeatModeOff.
 * Returns RLIB_EFAULT when app is NULL.
 */
int
rlib_x_get_autorepeat( RLXApp* app )
{
	XKeyboardState state;

	if ( app == NULL ) {
		return RLIB_EFAULT;
	}

	XGetKeyboardControl( app->display, &state );
	return state.global_auto_repeat;
}


/*!
 * Shows created window and start X event loop.
 * At each loop, this code will do the following.
 *
 * - Calls rlib_x_sync() and wait until the time for one frame passes.
 * - Reads X event if exists and save it into application context.
 * - Executes update function specified by rlib_x_update_func(). 
 *   If "drop frame" function is enabled and time for several frames 
 *   have passed, execute update function several times correspond to
 *   the passed time.
 *
 * - Executes drawing function specified by rlib_x_draw_func().
 *   This will be called only once at each loop 
 *   ( this is different from update function ).
 *   This function draws on back buffer at this point.
 *
 * - If debug function is enabled, draws some debug information
 *   onto the screen.
 *
 * - Copies the contents of back buffer to front buffer.
 *
 *
 * If rlib_x_stop is called, event loop will end, then cleanup function
 * will be executed.
 */
void
rlib_x_start( RLXApp* app )
{
	double accumulated_frame_count = 0.0;

	if ( app == NULL ) {
		return;
	}

	XMapWindow( app->display, app->frontbuffer );
	app->do_loop = 1;

	while ( app->do_loop ) {
		long elapsed_time = rlib_x_sync( app );
		int frame_count_for_debug_draw = 1;
		int frame_count = 1;

		if ( XPending( app->display ) ) {
			XNextEvent( app->display, &( app->event ) );

			if ( app->event.type == ClientMessage ) {
				if ( app->event.xclient.data.l[0] == app->atom_delete_window ) {
					rlib_x_stop( app, 0 );
					continue;
				}
			}
			else if ( app->event.type == FocusIn ) {
				rlib_x_autorepeat_off( app );
			}
			else if ( app->event.type == FocusOut ) {
				rlib_x_autorepeat_set( app, app->autorepeat_previous );
			}

			app->event_renewed = 1;
		}

		if ( app->drop_frame ) {
			accumulated_frame_count += 
				( double )( ( double )app->fps * ( double )elapsed_time / ( double )1000000 );
			frame_count_for_debug_draw = frame_count = ( int )accumulated_frame_count;
			accumulated_frame_count -= frame_count;
		}

		if ( app->update_func != NULL ) {
			for ( ; frame_count >= 1; frame_count -= 1 ) {
				app->update_func( app );
				app->event_renewed = 0;
			}
		}

		if ( app->draw_func != NULL ) {
			app->draw_func( app );
		}

		if ( app->debug ) {
			rlib_x_debug_draw( app, elapsed_time, frame_count_for_debug_draw );
		}

		rlib_x_flip( app );
	}

	rlib_x_cleanup( app );
}


static void
rlib_x_debug_draw( RLXApp* app, long elapsed_time, int frame_count )
{
#define RL_X_DEBUG_MESSAGE_LENGTH 128
#define RL_X_DEBUG_MESSAGE_WIDTH  600

	char debug_message[ RL_X_DEBUG_MESSAGE_LENGTH ];

	if ( app->drop_frame ) {
		sprintf( debug_message, 
			 "DropFrame = true,  Elapsed Time(usec) = %ld, Dropped Frames = %d, "
			 "Required FPS = %d, Current FPS = %ld",
			 elapsed_time, frame_count, app->fps, frame_count * 1000000 / elapsed_time
			);
	}
	else {
		sprintf( debug_message,
			 "DropFrame = false, Elapsed Time(usec) = %ld, Dropped Frames = 0, "
			 "Required FPS = %d, Current FPS = %ld",
			 elapsed_time, app->fps, frame_count * 1000000 / elapsed_time
			);
	}		

	XDrawImageString( 
		app->display, 
		app->backbuffer, 
		app->gc_debug, 
		app->width  / 2 - RL_X_DEBUG_MESSAGE_WIDTH / 2,
		app->height / 2,
		debug_message, 
		strlen( debug_message ) );
}


static void
rlib_x_flip( RLXApp* app )
{
	if ( app != NULL ) {
		XCopyArea(
			app->display,
			app->backbuffer,
			app->frontbuffer,
			app->gc,
			0,
			0,
			app->width,
			app->height,
			0,
			0 );
	}
}


/*!
 * When you want to stop event loop and quit, call this function.
 * Specified exit code will be saved into app->exitcode.
 */
void
rlib_x_stop( RLXApp* app, int exitcode )
{
	if ( app != NULL ) {
		app->exitcode = exitcode;
		app->do_loop  = 0;
	}
}


/*!
 * Cleanup function of X application tools.
 * This function will always be called when mail loop is going to stop by rlib_x_stop().
 * You should call this function manually when you're going to exit application
 * after rlib_x_initialize() is called and not by rlib_x_stop().
 */
void
rlib_x_cleanup( RLXApp* app )
{
	int i;

	if ( app == NULL ) {
		return;
	}

	/* 
	 * First, execute the user specified cleanup function.
	 * And then, execute cleanup procedures of the resources
	 * that RLXApp application context holds.
	 */

	if ( app->cleanup_func != NULL ) {
		app->cleanup_func( app->data_for_cleanup );
	}

	XFreeGC( app->display, app->gc );
	XFreeGC( app->display, app->gc_debug );
	XFreeGC( app->display, app->gc_pixmap );

	for ( i = 0; i < app->num_of_pixmaps; i ++ ) {
		Imlib_free_pixmap( app->imlib_data, app->pixmaps[i].body );
	}

	XFreePixmap( app->display, app->backbuffer );
	XDestroyWindow( app->display, app->frontbuffer );

	rlib_x_autorepeat_set( app, app->autorepeat_previous );
	XCloseDisplay( app->display );

	if ( app->debug ) {
		printf( "rlib_x_cleanup is called.\n" );
		printf( "Cleaning up of X resources have done.\n" );
	}
}


/*!
 * Gets exit code of event loop.
 */
int
rlib_x_get_exitcode( RLXApp* app, int* exitcode )
{
	if ( app == NULL || exitcode == NULL ) {
		return RLIB_EFAULT;
	}

	*exitcode = app->exitcode;
	return 0;
}


/*!
 * Gets drawable for the back buffer.
 */
int
rlib_x_get_drawable( RLXApp* app, Drawable* drawable )
{
	if ( app == NULL || drawable == NULL ) {
		return RLIB_EFAULT;
	}

	*drawable = app->backbuffer;
	return 0;
}


/*!
 * Waits until the time for one frame elapses from the previous
 * time when rlib_x_sync is called. Returns the elapsed time from 
 * the previous rlib_x_sync called time.
 */
long
rlib_x_sync( RLXApp* app )
{
	if ( app == NULL ) {
		return 0L;
	}

	while ( 1 ) {
		struct timeval current;
		struct timeval diff;

		long frame_time   = 1000000 / app->fps;
		long elapsed_time = rlib_x_get_elapsed_time( app, &current );
		long wait_time    = frame_time - elapsed_time;

		if ( elapsed_time >= frame_time ) {
			RLIB_TIMER_COPY( &( app->previous ), &current );
			return elapsed_time;
		}

		diff.tv_sec  = wait_time / 1000000;
		diff.tv_usec = wait_time - diff.tv_sec;

		select( 0, NULL, NULL, NULL, &diff );
	}
}


static long
rlib_x_get_elapsed_time( RLXApp* app, struct timeval* current_return )
{
	struct timeval current;
	struct timeval diff;
	long   elapsed_time;

	if ( ! timerisset( &( app->previous ) ) ) {
		gettimeofday( &current, NULL );
		RLIB_TIMER_COPY( &( app->previous ), &current );
		return 0L;
	}

	gettimeofday( &current, NULL );
	RLIB_TIMER_DIFF( &diff, &current, &(app->previous ) );
	elapsed_time = diff.tv_sec * 1000000 + diff.tv_usec;

	if ( current_return != NULL ) {
		RLIB_TIMER_COPY( current_return, &current );
	}

	return elapsed_time;
}


/*!
 * Enables debug function.
 * When debug function is enabled, some information will be
 * drawn on the screen ( "drop frame" status, required fps, current fps ).
 */
void
rlib_x_debug_on( RLXApp* app )
{
	if ( app != NULL ) {
		app->debug = 1;
	}
}


/*!
 * Disables debug function.
 */
void
rlib_x_debug_off( RLXApp* app )
{
	if ( app != NULL ) {
		app->debug = 0;
	}
}


/*!
 * Checks if debug function is enabled.
 * Returns 1 if enabled, 0 when not.
 */
int
rlib_x_is_debug_on( RLXApp* app )
{
	if ( app == NULL ) {
		return 0;
	}

	return app->debug;
}


/* End of RLxcore.c */
